from __future__ import annotations

import struct
from collections import OrderedDict
from textwrap import dedent
from typing import List, TYPE_CHECKING

from dissect.cstruct.bitbuffer import BitBuffer
from dissect.cstruct.expression import Expression
from dissect.cstruct.types import (
    Array,
    BytesInteger,
    CharType,
    Enum,
    EnumInstance,
    Field,
    Flag,
    FlagInstance,
    Instance,
    PackedType,
    Pointer,
    PointerInstance,
    Structure,
    Union,
    WcharType,
)

if TYPE_CHECKING:
    from dissect.cstruct import cstruct


class Compiler:
    """Compiler for cstruct structures. Creates somewhat optimized parsing code."""

    TYPES = (
        Structure,
        Pointer,
        Enum,
        Flag,
        Array,
        PackedType,
        CharType,
        WcharType,
        BytesInteger,
    )

    COMPILE_TEMPLATE = """
class {name}(Structure):
    def __init__(self, cstruct, structure, source=None):
        self.structure = structure
        self.source = source
        super().__init__(cstruct, structure.name, structure.fields, anonymous=structure.anonymous)

    def _read(self, stream):
        r = OrderedDict()
        sizes = {{}}
        bitreader = BitBuffer(stream, self.cstruct.endian)

{read_code}

        return Instance(self, r, sizes)

    def add_field(self, name, type_, offset=None):
        raise NotImplementedError("Can't add fields to a compiled structure")

    def __repr__(self):
        return '<Structure {name} +compiled>'
"""

    def __init__(self, cstruct: cstruct):
        self.cstruct = cstruct

    def compile(self, structure: Structure) -> Structure:
        if isinstance(structure, Union) or structure.align:
            return structure

        structure_name = structure.name

        try:
            # Generate struct class based on provided structure type
            source = self.gen_struct_class(structure_name, structure)
        except TypeError:
            return structure

        # Create code object that can be executed later on
        code_object = compile(
            source,
            f"<compiled {structure_name}>",
            "exec",
        )

        env = {
            "OrderedDict": OrderedDict,
            "Structure": Structure,
            "Instance": Instance,
            "Expression": Expression,
            "EnumInstance": EnumInstance,
            "FlagInstance": FlagInstance,
            "PointerInstance": PointerInstance,
            "BytesInteger": BytesInteger,
            "BitBuffer": BitBuffer,
            "struct": struct,
            "range": range,
        }

        exec(code_object, env)
        return env[structure_name](self.cstruct, structure, source)

    def gen_struct_class(self, name: str, structure: Structure) -> str:
        blocks = []
        classes = []
        cur_block = []
        read_size = 0
        prev_was_bits = False

        for field in structure.fields:
            field_type = self.cstruct.resolve(field.type)

            if not isinstance(field_type, self.TYPES):
                raise TypeError(f"Unsupported type for compiler: {field_type}")

            if isinstance(field_type, Structure) or (
                isinstance(field_type, Array) and isinstance(field_type.type, Structure)
            ):

                blocks.append(self.gen_read_block(read_size, cur_block))

                struct_read = "s = stream.tell()\n"
                if isinstance(field_type, Array):
                    num = field_type.count

                    if isinstance(num, Expression):
                        num = f"max(0, Expression(self.cstruct, '{num.expression}').evaluate(r))"

                    struct_read += dedent(
                        f"""
                        r['{field.name}'] = []
                        for _ in range({num}):
                            r['{field.name}'].append(self.lookup['{field.name}'].type.type._read(stream))
                        sizes['{field.name}'] = stream.tell() - s
                        """
                    )
                elif isinstance(field_type, Structure) and field_type.anonymous:
                    struct_read += dedent(
                        f"""
                        v = self.lookup["{field.name}"].type._read(stream)
                        r.update(v._values)
                        sizes.update(v._sizes)
                        """
                    )
                else:
                    struct_read += dedent(
                        f"""
                        r['{field.name}'] = self.lookup['{field.name}'].type._read(stream)
                        sizes['{field.name}'] = stream.tell() - s
                        """
                    )

                blocks.append(struct_read)
                read_size = 0
                cur_block = []
                continue

            if field.bits:
                blocks.append(self.gen_read_block(read_size, cur_block))
                if isinstance(field_type, Enum):
                    bitfield_read = dedent(
                        f"""
                        r['{field.name}'] = self.cstruct.{field.type.name}(
                            bitreader.read(self.cstruct.{field.type.type.name}, {field.bits})
                        )
                        """
                    )
                else:
                    bitfield_read = f"r['{field.name}'] = bitreader.read(self.cstruct.{field.type.name}, {field.bits})"
                blocks.append(bitfield_read)

                read_size = 0
                cur_block = []
                prev_was_bits = True
                continue

            if prev_was_bits:
                blocks.append("bitreader.reset()")
                prev_was_bits = False

            try:
                count = len(field_type)
                read_size += count
                cur_block.append(field)
            except TypeError:
                if cur_block:
                    blocks.append(self.gen_read_block(read_size, cur_block))

                blocks.append(self.gen_dynamic_block(field))
                read_size = 0
                cur_block = []

        if len(cur_block):
            blocks.append(self.gen_read_block(read_size, cur_block))

        read_code = "\n\n".join(blocks)
        read_code = "\n".join(["    " * 2 + line for line in read_code.split("\n")])

        classes.append(self.COMPILE_TEMPLATE.format(name=name, read_code=read_code))
        return "\n\n".join(classes)

    def gen_read_block(self, size: int, block: List[str]) -> str:
        template = dedent(
            f"""
            buf = stream.read({size})
            if len(buf) != {size}: raise EOFError()
            data = struct.unpack(self.cstruct.endian + '{{}}', buf)
            {{}}
            """
        )

        read_code = []
        fmt = []

        cur_type = None
        cur_count = 0

        buf_offset = 0
        data_offset = 0

        for field in block:
            field_type = self.cstruct.resolve(field.type)
            read_type = field_type

            count = 1
            data_count = 1

            if isinstance(read_type, (Enum, Flag)):
                read_type = read_type.type
            elif isinstance(read_type, Pointer):
                read_type = self.cstruct.pointer

            if isinstance(field_type, Array):
                count = read_type.count
                data_count = count
                read_type = read_type.type

                if isinstance(read_type, (Enum, Flag)):
                    read_type = read_type.type
                elif isinstance(read_type, Pointer):
                    read_type = self.cstruct.pointer

                if isinstance(read_type, (CharType, WcharType, BytesInteger)):
                    read_slice = f"{buf_offset}:{buf_offset + + (count * read_type.size)}"
                else:
                    read_slice = f"{data_offset}:{data_offset + count}"
            elif isinstance(read_type, CharType):
                read_slice = f"{buf_offset}:{buf_offset + 1}"
            elif isinstance(read_type, (WcharType, BytesInteger)):
                read_slice = f"{buf_offset}:{buf_offset + read_type.size}"
            else:
                read_slice = str(data_offset)

            if not cur_type:
                if isinstance(read_type, PackedType):
                    cur_type = read_type.packchar
                else:
                    cur_type = "x"

            if isinstance(read_type, (PackedType, CharType, WcharType, BytesInteger, Enum, Flag)):
                char_count = count

                if isinstance(read_type, (CharType, WcharType, BytesInteger)):
                    data_count = 0
                    pack_char = "x"
                    char_count *= read_type.size
                else:
                    pack_char = read_type.packchar

                if cur_type != pack_char:
                    fmt.append(f"{cur_count}{cur_type}")
                    cur_count = 0

                cur_count += char_count
                cur_type = pack_char

            if isinstance(read_type, BytesInteger):
                getter = "BytesInteger.parse(buf[{slice}], {size}, {count}, {signed}, self.cstruct.endian){data_slice}"

                getter = getter.format(
                    slice=read_slice,
                    size=read_type.size,
                    count=count,
                    signed=read_type.signed,
                    data_slice="[0]" if count == 1 else "",
                )
            elif isinstance(read_type, (CharType, WcharType)):
                getter = f"buf[{read_slice}]"

                if isinstance(read_type, WcharType):
                    getter += ".decode('utf-16-le' if self.cstruct.endian == '<' else 'utf-16-be')"
            else:
                getter = f"data[{read_slice}]"

            if isinstance(field_type, (Enum, Flag)):
                enum_type = field_type.__class__.__name__
                getter = f"{enum_type}Instance(self.cstruct.{field_type.name}, {getter})"
            elif isinstance(field_type, Array) and isinstance(field_type.type, (Enum, Flag)):
                enum_type = field_type.type.__class__.__name__
                getter = f"[{enum_type}Instance(self.cstruct.{field_type.type.name}, d) for d in {getter}]"
            elif isinstance(field_type, Pointer):
                getter = f"PointerInstance(self.cstruct.{field_type.type.name}, stream, {getter}, r)"
            elif isinstance(field_type, Array) and isinstance(field_type.type, Pointer):
                getter = f"[PointerInstance(self.cstruct.{field_type.type.name}, stream, d, r) for d in {getter}]"
            elif isinstance(field_type, Array) and isinstance(read_type, PackedType):
                getter = f"list({getter})"

            read_code.append(f"r['{field.name}'] = {getter}")
            read_code.append(f"sizes['{field.name}'] = {count * read_type.size}")

            data_offset += data_count
            buf_offset += count * read_type.size

        if cur_count:
            fmt.append(f"{cur_count}{cur_type}")

        return template.format("".join(fmt), "\n".join(read_code))

    def gen_dynamic_block(self, field: Field) -> str:
        if not isinstance(field.type, Array):
            raise TypeError(f"Only Array can be dynamic, got {field.type!r}")

        field_type = self.cstruct.resolve(field.type.type)
        reader = None

        if isinstance(field_type, (Enum, Flag)):
            field_type = field_type.type

        if not field.type.count:  # Null terminated
            if isinstance(field_type, PackedType):
                reader = dedent(
                    f"""
                    t = []
                    while True:
                        d = stream.read({field_type.size})
                        if len(d) != {field_type.size}: raise EOFError()
                        v = struct.unpack(self.cstruct.endian + '{field_type.packchar}', d)[0]
                        if v == 0: break
                        t.append(v)
                    """
                )

            elif isinstance(field_type, (CharType, WcharType)):
                null = "\\x00" * field_type.size
                reader = dedent(
                    f"""
                    t = []
                    while True:
                        c = stream.read({field_type.size})
                        if len(c) != {field_type.size}: raise EOFError()
                        if c == b'{null}': break
                        t.append(c)
                    t = b''.join(t)"""  # It's important there's no newline here because of the optional decode
                )

                if isinstance(field_type, WcharType):
                    reader += ".decode('utf-16-le' if self.cstruct.endian == '<' else 'utf-16-be')"
            elif isinstance(field_type, BytesInteger):
                reader = dedent(
                    f"""
                    t = []
                    while True:
                        d = stream.read({field_type.size})
                        if len(d) != {field_type.size}: raise EOFError()
                        v = BytesInteger.parse(d, {field_type.size}, 1, {field_type.signed}, self.cstruct.endian)
                        if v == 0: break
                        t.append(v)
                    """
                )

            if isinstance(field_type, (Enum, Flag)):
                enum_type = field_type.__class__.__name__
                reader += f"\nt = [{enum_type}Instance(self.cstruct.{field_type.name}, d) for d in t]"

            if not reader:
                raise TypeError(f"Couldn't compile a reader for array {field!r}, {field_type!r}.")

            return f"s = stream.tell()\n{reader}\nr['{field.name}'] = t\nsizes['{field.name}'] = stream.tell() - s"

        expr_read = dedent(
            f"""
            dynsize = max(0, Expression(self.cstruct, "{field.type.count.expression}").evaluate(r))
            buf = stream.read(dynsize * {field_type.size})
            if len(buf) != dynsize * {field_type.size}: raise EOFError()
            r['{field.name}'] = {{reader}}
            sizes['{field.name}'] = dynsize * {field_type.size}
            """
        )

        if isinstance(field_type, PackedType):
            reader = f"list(struct.unpack(self.cstruct.endian + f'{{dynsize}}{field_type.packchar}', buf))"
        elif isinstance(field_type, (CharType, WcharType)):
            reader = "buf"
            if isinstance(field_type, WcharType):
                reader += ".decode('utf-16-le' if self.cstruct.endian == '<' else 'utf-16-be')"
        elif isinstance(field_type, BytesInteger):
            reader = f"BytesInteger.parse(buf, {field_type.size}, dynsize, {field_type.signed}, self.cstruct.endian)"

        if isinstance(field_type, (Enum, Flag)):
            enum_type = field_type.__class__.__name__
            reader += f"[{enum_type}Instance(self.cstruct.{field_type.name}, d) for d in {reader}]"

        return expr_read.format(reader=reader, size=None)



from dissect import cstruct

from .utils import verify_compiled


def test_union():
    cdef = """
    union test {
        uint32 a;
        char   b[8];
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=False)

    assert len(cs.test) == 8

    buf = b"zomgbeef"
    obj = cs.test(buf)

    assert obj.a == 0x676D6F7A
    assert obj.b == b"zomgbeef"

    assert obj.dumps() == buf
    assert cs.test().dumps() == b"\x00\x00\x00\x00\x00\x00\x00\x00"


def test_union_nested(compiled):
    cdef = """
    struct test {
        char magic[4];
        union {
            struct {
                uint32 a;
                uint32 b;
            } a;
            struct {
                char   b[8];
            } b;
        } c;
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    assert len(cs.test) == 12

    buf = b"zomgholybeef"
    obj = cs.test(buf)

    assert obj.magic == b"zomg"
    assert obj.c.a.a == 0x796C6F68
    assert obj.c.a.b == 0x66656562
    assert obj.c.b.b == b"holybeef"

    assert obj.dumps() == buf


def test_union_anonymous(compiled):
    cdef = """
    typedef struct test
    {
        union
        {
            uint32 a;
            struct
            {
                char b[3];
                char c;
            };
        };
        uint32 d;
    }
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    buf = b"\x01\x01\x02\x02\x03\x03\x04\x04"
    obj = cs.test(buf)

    assert obj.a == 0x02020101
    assert obj.b == b"\x01\x01\x02"
    assert obj.c == b"\x02"
    assert obj.d == 0x04040303
    assert obj.dumps() == buf



from io import BytesIO

import pytest

from dissect import cstruct
from dissect.cstruct.bitbuffer import BitBuffer


def test_bitbuffer_read():
    cs = cstruct.cstruct()

    bb = BitBuffer(BytesIO(b"\xff"), "<")
    assert bb.read(cs.uint8, 8) == 0b11111111

    bb = BitBuffer(BytesIO(b"\xf0"), "<")
    assert bb.read(cs.uint8, 4) == 0b0000
    assert bb.read(cs.uint8, 4) == 0b1111

    bb = BitBuffer(BytesIO(b"\xf0"), ">")
    assert bb.read(cs.uint8, 4) == 0b1111
    assert bb.read(cs.uint8, 4) == 0b0000

    bb = BitBuffer(BytesIO(b"\xff\x00"), "<")
    assert bb.read(cs.uint16, 12) == 0b11111111
    assert bb.read(cs.uint16, 4) == 0b0

    bb = BitBuffer(BytesIO(b"\xff\x00"), ">")
    assert bb.read(cs.uint16, 12) == 0b000000001111
    assert bb.read(cs.uint16, 4) == 0b1111

    bb = BitBuffer(BytesIO(b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"), "<")
    with pytest.raises(ValueError) as exc:
        assert bb.read(cs.uint32, 160)
    assert str(exc.value) == "Reading straddled bits is unsupported"

    assert bb.read(cs.uint32, 30) == 0b111111111111111111111111111111
    assert bb.read(cs.uint32, 2) == 0b11
    assert bb.read(cs.uint32, 32) == 0b11111111111111111111111111111111



from dissect.cstruct.types.base import Array, BaseType, RawType
from dissect.cstruct.types.bytesinteger import BytesInteger
from dissect.cstruct.types.chartype import CharType
from dissect.cstruct.types.enum import Enum, EnumInstance
from dissect.cstruct.types.flag import Flag, FlagInstance
from dissect.cstruct.types.instance import Instance
from dissect.cstruct.types.packedtype import PackedType
from dissect.cstruct.types.pointer import Pointer, PointerInstance
from dissect.cstruct.types.structure import Field, Structure, Union
from dissect.cstruct.types.voidtype import VoidType
from dissect.cstruct.types.wchartype import WcharType

__all__ = [
    "Array",
    "BaseType",
    "BytesInteger",
    "CharType",
    "Enum",
    "EnumInstance",
    "Field",
    "Flag",
    "FlagInstance",
    "Instance",
    "PackedType",
    "Pointer",
    "PointerInstance",
    "RawType",
    "Structure",
    "Union",
    "VoidType",
    "WcharType",
]



from __future__ import annotations

import struct
from typing import BinaryIO, List, TYPE_CHECKING

from dissect.cstruct.types import RawType

if TYPE_CHECKING:
    from dissect.cstruct import cstruct


class PackedType(RawType):
    """Implements a packed type that uses Python struct packing characters."""

    def __init__(self, cstruct: cstruct, name: str, size: int, packchar: str, alignment: int = None):
        super().__init__(cstruct, name, size, alignment)
        self.packchar = packchar

    def _read(self, stream: BinaryIO) -> int:
        return self._read_array(stream, 1)[0]

    def _read_array(self, stream: BinaryIO, count: int) -> List[int]:
        length = self.size * count
        data = stream.read(length)
        fmt = self.cstruct.endian + str(count) + self.packchar

        if len(data) != length:
            raise EOFError(f"Read {len(data)} bytes, but expected {length}")

        return list(struct.unpack(fmt, data))

    def _read_0(self, stream: BinaryIO) -> List[int]:
        byte_array = []
        while True:
            bytes_stream = stream.read(self.size)
            unpacked_struct = struct.unpack(self.cstruct.endian + self.packchar, bytes_stream)[0]

            if unpacked_struct == 0:
                break

            byte_array.append(unpacked_struct)

        return byte_array

    def _write(self, stream: BinaryIO, data: int) -> int:
        return self._write_array(stream, [data])

    def _write_array(self, stream: BinaryIO, data: List[int]) -> int:
        fmt = self.cstruct.endian + str(len(data)) + self.packchar
        return stream.write(struct.pack(fmt, *data))

    def _write_0(self, stream: BinaryIO, data: List[int]) -> int:
        return self._write_array(stream, data + [0])

    def default(self) -> int:
        return 0

    def default_array(self, count: int) -> List[int]:
        return [0] * count



import pytest

from dissect import cstruct

from .utils import verify_compiled


def test_bitfield(compiled):
    cdef = """
    struct test {
        uint16  a:4;
        uint16  b:4;
        uint16  c:4;
        uint16  d:4;
        uint32  e;
        uint16  f:2;
        uint16  g:3;
        uint32  h;
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)
    assert len(cs.test) == 12

    buf = b"\x12\x34\xff\x00\x00\x00\x1f\x00\x01\x00\x00\x00"
    obj = cs.test(buf)

    assert obj.a == 0b10
    assert obj.b == 0b01
    assert obj.c == 0b100
    assert obj.d == 0b011
    assert obj.e == 0xFF
    assert obj.f == 0b11
    assert obj.g == 0b111
    assert obj.h == 1
    assert obj.dumps() == buf


def test_bitfield_consecutive(compiled):
    cdef = """
    struct test {
        uint16  a:4;
        uint16  b:4;
        uint16  c:4;
        uint16  d:4;
        uint16  e:16;
        uint16  _pad1;
        uint16  f:2;
        uint16  g:3;
        uint16  h:11;
        uint16  _pad2;
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)
    assert len(cs.test) == 10

    buf = b"\x12\x34\xff\x00\x00\x00\x1f\x01\x00\x00"
    obj = cs.test(buf)

    assert obj.a == 0b10
    assert obj.b == 0b01
    assert obj.c == 0b100
    assert obj.d == 0b011
    assert obj.e == 0xFF
    assert obj.f == 0b11
    assert obj.g == 0b111
    assert obj.h == 0b1000
    assert obj.dumps() == buf


def test_struct_after_bitfield(compiled):
    cdef = """
    struct test {
        uint16  a:4;
        uint16  b:4;
        uint32  c;
        struct {
            uint32 d;
        } nested1;
        uint16  e:4;
        uint16  f:4;
        struct {
            uint32 g;
        } nested2;
        uint32  h;
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)
    assert len(cs.test) == 20

    buf = b"\x12\x00\xff\x00\x00\x00\x01\x00\x00\x00\x12\x00\x02\x00\x00\x00\xfe\x00\x00\x00"
    obj = cs.test(buf)

    assert obj.a == 0b10
    assert obj.b == 0b01
    assert obj.c == 0xFF
    assert obj.nested1.d == 0x01
    assert obj.e == 0b10
    assert obj.f == 0b01
    assert obj.nested2.g == 0x02
    assert obj.h == 0xFE
    assert obj.dumps() == buf


def test_bitfield_be(compiled):
    cdef = """
    struct test {
        uint16  a:4;
        uint16  b:4;
        uint16  c:4;
        uint16  d:4;
        uint32  e;
        uint16  f:2;
        uint16  g:3;
        uint16  h:4;
        uint32  i;
    };
    """
    cs = cstruct.cstruct(endian=">")
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)
    assert len(cs.test) == 12

    buf = b"\x12\x34\x00\x00\x00\xff\x1f\x00\x00\x00\x00\x01"
    obj = cs.test(buf)

    assert obj.a == 0b01
    assert obj.b == 0b10
    assert obj.c == 0b011
    assert obj.d == 0b100
    assert obj.e == 0xFF
    assert obj.f == 0
    assert obj.g == 0b11
    assert obj.h == 0b1110
    assert obj.i == 1
    assert obj.dumps() == buf


def test_bitfield_straddle(compiled):
    cdef = """
    struct test {
        uint16  a:12;
        uint16  b:12;
        uint16  c:8;
        uint32  d;
        uint16  e:2;
        uint16  f:3;
        uint32  g;
    };
    """
    cs = cstruct.cstruct()

    with pytest.raises(ValueError) as exc:
        cs.load(cdef, compiled=compiled)

    assert str(exc.value) == "Straddled bit fields are unsupported"


def test_bitfield_write(compiled):
    cdef = """
    struct test {
        uint16  a:1;
        uint16  b:1;
        uint32  c;
        uint16  d:2;
        uint16  e:3;
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    obj = cs.test()
    obj.a = 0b1
    obj.b = 0b1
    obj.c = 0xFF
    obj.d = 0b11
    obj.e = 0b111

    assert obj.dumps() == b"\x03\x00\xff\x00\x00\x00\x1f\x00"


def test_bitfield_write_be(compiled):
    cdef = """
    struct test {
        uint16  a:1;
        uint16  b:1;
        uint32  c;
        uint16  d:2;
        uint16  e:3;
    };
    """
    cs = cstruct.cstruct(endian=">")
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    obj = cs.test()
    obj.a = 0b1
    obj.b = 0b1
    obj.c = 0xFF
    obj.d = 0b11
    obj.e = 0b111

    assert obj.dumps() == b"\xc0\x00\x00\x00\x00\xff\xf8\x00"


def test_bitfield_with_enum_or_flag(compiled):
    cdef = """
    flag Flag8 : uint8 {
        A = 1,
        B = 2
    };

    flag Flag16 : uint16 {
        A = 1,
        B = 2
    };

    struct test {
        uint16  a:1;
        uint16  b:1;
        Flag16  c:2;
        Flag8   d:4;
    };
    """
    cs = cstruct.cstruct(endian=">")
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    buf = b"\xf0\x00\x30"
    obj = cs.test(buf)

    assert obj.a == 0b1
    assert obj.b == 0b1
    assert obj.c == cs.Flag16.A | cs.Flag16.B
    assert obj.d == cs.Flag8.A | cs.Flag8.B

    assert obj.dumps() == buf



#!/usr/bin/env python
# flake8: noqa
import zlib
import binascii
from io import BytesIO

from dissect.cstruct import cstruct

cdef = """
struct SECURITY_DESCRIPTOR {
    uint8   Revision;
    uint8   Sbz1;
    uint16  Control;
    uint32  OffsetOwner;
    uint32  OffsetGroup;
    uint32  OffsetSacl;
    uint32  OffsetDacl;
};

struct LDAP_SID_IDENTIFIER_AUTHORITY {
    char    Value[6];
};

struct LDAP_SID {
    uint8   Revision;
    uint8   SubAuthorityCount;
    LDAP_SID_IDENTIFIER_AUTHORITY   IdentifierAuthority;
    uint32  SubAuthority[SubAuthorityCount];
};

struct ACL {
    uint8   AclRevision;
    uint8   Sbz1;
    uint16  AclSize;
    uint16  AceCount;
    uint16  Sbz2;
    char    Data[AclSize - 8];
};

struct ACE {
    uint8   AceType;
    uint8   AceFlags;
    uint16  AceSize;
    char    Data[AceSize - 4];
};

struct ACCESS_ALLOWED_ACE {
    uint16  Mask;
    LDAP_SID Sid;
};

struct ACCESS_ALLOWED_OBJECT_ACE {
    uint32  Mask;
    uint32  Flags;
    char    ObjectType[Flags & 1 * 16];
    char    InheritedObjectType[Flags & 2 * 8];
    LDAP_SID Sid;
};
"""
c_secd = cstruct()
c_secd.load(cdef, compiled=True)


class SecurityDescriptor:
    def __init__(self, fh):
        self.fh = fh
        self.descriptor = c_secd.SECURITY_DESCRIPTOR(fh)

        self.owner_sid = b""
        self.group_sid = b""
        self.sacl = b""
        self.dacl = b""

        if self.descriptor.OffsetOwner != 0:
            fh.seek(self.descriptor.OffsetOwner)
            self.owner_sid = LdapSid(fh=fh)

        if self.descriptor.OffsetGroup != 0:
            fh.seek(self.descriptor.OffsetGroup)
            self.group_sid = LdapSid(fh=fh)

        if self.descriptor.OffsetSacl != 0:
            fh.seek(self.descriptor.OffsetSacl)
            self.sacl = ACL(fh)

        if self.descriptor.OffsetDacl != 0:
            fh.seek(self.descriptor.OffsetDacl)
            self.dacl = ACL(fh)


class LdapSid:
    def __init__(self, fh=None, in_obj=None):
        if fh:
            self.fh = fh
            self.ldap_sid = c_secd.LDAP_SID(fh)
        else:
            self.ldap_sid = in_obj

    def __repr__(self):
        return "S-{}-{}-{}".format(
            self.ldap_sid.Revision,
            bytearray(self.ldap_sid.IdentifierAuthority.Value)[5],
            "-".join(["{}".format(v) for v in self.ldap_sid.SubAuthority]),
        )


class ACL:
    def __init__(self, fh):
        self.fh = fh
        self.acl = c_secd.ACL(fh)
        self.aces = []

        buf = BytesIO(self.acl.Data)
        for i in range(self.acl.AceCount):
            self.aces.append(ACE(buf))


class ACCESS_ALLOWED_ACE:
    def __init__(self, fh):
        self.fh = fh
        self.data = c_secd.ACCESS_ALLOWED_ACE(fh)
        self.sid = LdapSid(in_obj=self.data.Sid)


class ACCESS_DENIED_ACE(ACCESS_ALLOWED_ACE):
    pass


class ACCESS_ALLOWED_OBJECT_ACE:
    # Flag constants
    ACE_OBJECT_TYPE_PRESENT = 0x01
    ACE_INHERITED_OBJECT_TYPE_PRESENT = 0x02

    # ACE type specific mask constants
    # Note that while not documented, these also seem valid
    # for ACCESS_ALLOWED_ACE types
    ADS_RIGHT_DS_CONTROL_ACCESS = 0x00000100
    ADS_RIGHT_DS_CREATE_CHILD = 0x00000001
    ADS_RIGHT_DS_DELETE_CHILD = 0x00000002
    ADS_RIGHT_DS_READ_PROP = 0x00000010
    ADS_RIGHT_DS_WRITE_PROP = 0x00000020
    ADS_RIGHT_DS_SELF = 0x00000008

    def __init__(self, fh):
        self.fh = fh
        self.data = c_secd.ACCESS_ALLOWED_OBJECT_ACE(fh)
        self.sid = LdapSid(in_obj=self.data.Sid)


class ACCESS_DENIED_OBJECT_ACE(ACCESS_ALLOWED_OBJECT_ACE):
    pass


class ACCESS_MASK:
    """
    ACCESS_MASK as described in 2.4.3
    https://msdn.microsoft.com/en-us/library/cc230294.aspx
    """

    # Flag constants
    GENERIC_READ = 0x80000000
    GENERIC_WRITE = 0x04000000
    GENERIC_EXECUTE = 0x20000000
    GENERIC_ALL = 0x10000000
    MAXIMUM_ALLOWED = 0x02000000
    ACCESS_SYSTEM_SECURITY = 0x01000000
    SYNCHRONIZE = 0x00100000
    WRITE_OWNER = 0x00080000
    WRITE_DACL = 0x00040000
    READ_CONTROL = 0x00020000
    DELETE = 0x00010000

    def __init__(self, mask):
        self.mask = mask

    def has_priv(self, priv):
        return (self.mask & priv) == priv

    def set_priv(self, priv):
        self.mask |= priv

    def remove_priv(self, priv):
        self.mask ^= priv


class ACE:
    CONTAINER_INHERIT_ACE = 0x01
    FAILED_ACCESS_ACE_FLAG = 0x80
    INHERIT_ONLY_ACE = 0x08
    INHERITED_ACE = 0x10
    NO_PROPAGATE_INHERIT_ACE = 0x04
    OBJECT_INHERIT_ACE = 0x01
    SUCCESSFUL_ACCESS_ACE_FLAG = 0x04

    def __init__(self, fh):
        self.fh = fh
        self.ace = c_secd.ACE(fh)
        self.acedata = None
        buf = BytesIO(self.ace.Data)
        if self.ace.AceType == 0x00:
            # ACCESS_ALLOWED_ACE
            self.acedata = ACCESS_ALLOWED_ACE(buf)
        elif self.ace.AceType == 0x05:
            # ACCESS_ALLOWED_OBJECT_ACE
            self.acedata = ACCESS_ALLOWED_OBJECT_ACE(buf)
        elif self.ace.AceType == 0x01:
            # ACCESS_DENIED_ACE
            self.acedata = ACCESS_DENIED_ACE(buf)
        elif self.ace.AceType == 0x06:
            # ACCESS_DENIED_OBJECT_ACE
            self.acedata = ACCESS_DENIED_OBJECT_ACE(buf)
        # else:
        #     print 'Unsupported type %d' % self.ace.AceType

        if self.acedata:
            self.mask = ACCESS_MASK(self.acedata.data.Mask)

    def __repr__(self):
        return repr(self.ace)


if __name__ == "__main__":
    d = BytesIO(
        zlib.decompress(
            binascii.unhexlify(
                "789c636410e949916060680062110606861e206661a8606002d2ec51160c0a409a1988f759f37dfe30"
                "ffa2e03620e70773fac1a555d3f63fe3bd20b8a89561158381e723464606100093c8faf693a08f85e1"
                "867809906665f080eb175050885f5a7941708202c301ff4357ceb7a1eb6705eb671505d9354365a9b7"
                "e91ff127ae3ed9200f209b135089aa8f5c7382a964ce85fd5c7659421704172400f5e518bf25cd1c0b"
                "060e68a0b94f7d5c24517d5170ed7b60f8dcb87a96587d1deecb3eb3055f145c7914122f84f429c022"
                "c949cce780c205c1e5194077e6b16a12d26700d5574fa4ff58c1fa74c06104d2b7025d1f13441fc83d"
                "6a50b50250b5b21b57baad4b8872d8fee27f57c495b04bc86a0dc06a35c00a416a83a58a56ebcb5d10"
                "9c210934d7216836723a04a9c319be1075ac5c68ea30c2134d1d2c1cda766c2df79a725170dd5e4c75"
                "0c0c2a0c57189919f0852948cd7f467e026a24206a90fcaf02668b40c215c54e118629404d30316ea8"
                "18583f544c084c20f2b42987ce8e17323ece3b3e843c929ab7e12646da45f6bb10240d80f4bd09489a"
                "bffbbeae7383e263e5eb9f184ee0d527e5010e30662c71b58bc8b454c6826a4ec1544d865c1544da25"
                "d51c76a83f184f969e7c9593efdd2c79added5652a1bb1e694b040c283119a1644ce6878988bec719d"
                "cdbe369f312e5e03affd48fa882d2bd0f5cd21531f46994da43e92c21749df05b92d313e7c407d6dc4"
                "e983e57f62cba642347de24b366f0c7d33d171b3fa763de3c75d9f08c51f4c9febcdaa59c1a72e0aee"
                "bec0d0703cade000b1fa32484c2f307d9d5d9af233eeed70df7a365cebaaad5a1db1fa26569267df42"
                "952bf149487514217d0a507d6c44fa2f834c7da568fab8484c67307d7cdf4aec8aed8065f601fc7520"
                "ba3b89d587ee4e62d319ba7dc4ea43b74f8ac87847b78f587de8f6c931cd9a15ed050c97c3f8c3053d"
                "bd281c9cc674cb817078a2c79fda4bdfc9f3365c125cc506ac9bfbdede20d63e770bd3b89ccffb3c96"
                "7ff7cc08610ab021565fc029ebde3a59607e68242d3fa0b719898d0762f5a1c7430863ef931f7b2e0a"
                "b63301c365674200d1faf4a3d575679cf5ddb0f6a929a34bdc6f62fd1742a6ff88d587eece4432ed23"
                "561fba7dc496d7e8e9b390c8f2135d5ff97383905787551c26e8dd7d3849c12d93d878a84c48c86fb2"
                "92f6e9fb7de7c44959b57fc4eaab22d37fc4b6e9d1ed6b7ae595786cf505df851ed7aa228f6b8811ab"
                "afa5d239ede8e17ecf4d99332a24df7b7813ab0fa3cd4d64bc7796dc5ff1ea28b07e3f8dbf7e474f67"
                "c4eac3b08fc8fa1dc33e22f5a1db37ebff8f0f130509a74f74fb88d5871e0fc4ea4377e7dcbc033a75"
                "f9594e1daa4ca2f7c4f30489b58fd8f60bbaff88d587eece1d8f938db2ff26f8b47f32d9bb6a6ee66b"
                "62ddb9872f59e3aa23b0fe3b485a3b84587de8ee3c406439889edfd1c72c88d577f9f8969af6f60d4e"
                "5b2c6c62dd675d9623142eb0be32ff35f7e809099b1ce69b6bf9bee83748c6a58f0fd4b797b2607060"
                "6001777f49ea5721e923b5bf02ee2303e53678763036aebc24b8f23ffe76480699fa4aa1ee04850ac9"
                "fe1342e823a9df88641f49fd46247d24f51b91f491d62f47e823a9df2880180b5b96cb34db8637c6ad"
                "fb5de0cceb6231bbf08e694941f43163d187633c83196c9810c9fae0e32702507db9f38e6d3fae7349"
                "b0d58f61c1c9e66f1cb8f4713240dc894b1f467c12a90f235e90f5218d0f4dae967ee51177d56d4f8e"
                "ef97bfcbbb4cf18ff3e880c79fb0a66ba4f12b2d2654b5187ec0a316c3dde86a8534708f0540dd2ac2"
                "80aa0ea3ce85aae386aa8395614f0f57d8cffabed76dc18eb91282b9772a51e316a88e11a2ee9efbb3"
                "89f9370bbca7865fb3fff2f9cc0d94313a2115c8181d9e7285909a52a81a7c6387ec4c84d5f083d548"
                "30808c430f4790f85eb4714705201b5f7ec2270700bed3f0c5"
            )
        )
    )

    sc = SecurityDescriptor(d)

    # print sc.descriptor
    # print sc.owner_sid
    # print sc.group_sid
    # print sc.sacl.acl
    # print sc.sacl.aces
    # print sc.dacl.acl
    # print sc.dacl.aces

    for i in sc.dacl.aces:
        if i.ace.AceType == 0x05:
            print(i.acedata.sid)



from __future__ import annotations

import operator
from typing import Any, BinaryIO, Callable, Dict, Union, TYPE_CHECKING

from dissect.cstruct.exceptions import NullPointerDereference
from dissect.cstruct.types import Array, BaseType, RawType

if TYPE_CHECKING:
    from dissect.cstruct import cstruct


class Pointer(RawType):
    """Implements a pointer to some other type."""

    def __init__(self, cstruct: cstruct, target: BaseType):
        self.cstruct = cstruct
        self.type = target
        super().__init__(cstruct, "pointer", self.cstruct.pointer.size, self.cstruct.pointer.alignment)

    def __repr__(self) -> str:
        return f"<Pointer {self.type}>"

    def _read(self, stream: BinaryIO, ctx: Dict[str, Any]) -> PointerInstance:
        addr = self.cstruct.pointer(stream)
        return PointerInstance(self.type, stream, addr, ctx)

    def _write(self, stream: BinaryIO, data: Union[int, PointerInstance]):
        if isinstance(data, PointerInstance):
            data = data._addr

        if not isinstance(data, int):
            raise TypeError("Invalid pointer data")

        return self.cstruct.pointer._write(stream, data)


class PointerInstance:
    """Like the Instance class, but for structures referenced by a pointer."""

    def __init__(self, type_: BaseType, stream: BinaryIO, addr: int, ctx: Dict[str, Any]):
        self._stream = stream
        self._type = type_
        self._addr = addr
        self._ctx = ctx
        self._value = None

    def __repr__(self) -> str:
        return f"<Pointer {self._type} @ 0x{self._addr:x}>"

    def __str__(self) -> str:
        return str(self.dereference())

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.dereference(), attr)

    def __int__(self) -> int:
        return self._addr

    def __nonzero__(self) -> bool:
        return self._addr != 0

    def __addr_math(self, other: Union[int, PointerInstance], op: Callable[[int, int], int]) -> PointerInstance:
        if isinstance(other, PointerInstance):
            other = other._addr

        return PointerInstance(self._type, self._stream, op(self._addr, other), self._ctx)

    def __add__(self, other: Union[int, PointerInstance]) -> PointerInstance:
        return self.__addr_math(other, operator.__add__)

    def __sub__(self, other: Union[int, PointerInstance]) -> PointerInstance:
        return self.__addr_math(other, operator.__sub__)

    def __mul__(self, other: Union[int, PointerInstance]) -> PointerInstance:
        return self.__addr_math(other, operator.__mul__)

    def __floordiv__(self, other: Union[int, PointerInstance]) -> PointerInstance:
        return self.__addr_math(other, operator.__floordiv__)

    def __mod__(self, other: Union[int, PointerInstance]) -> PointerInstance:
        return self.__addr_math(other, operator.__mod__)

    def __pow__(self, other: Union[int, PointerInstance]) -> PointerInstance:
        return self.__addr_math(other, operator.__pow__)

    def __lshift__(self, other: Union[int, PointerInstance]) -> PointerInstance:
        return self.__addr_math(other, operator.__lshift__)

    def __rshift__(self, other: Union[int, PointerInstance]) -> PointerInstance:
        return self.__addr_math(other, operator.__rshift__)

    def __and__(self, other: Union[int, PointerInstance]) -> PointerInstance:
        return self.__addr_math(other, operator.__and__)

    def __xor__(self, other: Union[int, PointerInstance]) -> PointerInstance:
        return self.__addr_math(other, operator.__xor__)

    def __or__(self, other: Union[int, PointerInstance]) -> PointerInstance:
        return self.__addr_math(other, operator.__or__)

    def __eq__(self, other: Union[int, PointerInstance]) -> bool:
        if isinstance(other, PointerInstance):
            other = other._addr

        return self._addr == other

    def dereference(self) -> Any:
        if self._addr == 0:
            raise NullPointerDereference()

        if self._value is None:
            # Read current position of file read/write pointer
            position = self._stream.tell()
            # Reposition the file read/write pointer
            self._stream.seek(self._addr)

            if isinstance(self._type, Array):
                value = self._type._read(self._stream, self._ctx)
            else:
                value = self._type._read(
                    self._stream,
                )

            self._stream.seek(position)
            self._value = value

        return self._value



from __future__ import annotations

from typing import BinaryIO, List, Tuple, Union

from dissect.cstruct.types import Enum, EnumInstance


class Flag(Enum):
    """Implements a Flag type.

    Flags can be made using any type. The API for accessing flags and their
    values is very similar to Python 3 native flags.

    Example:
        When using the default C-style parser, the following syntax is supported:

            flag <name> [: <type>] {
                <values>
            };

        For example, a flag that has A=1, B=4 and C=8 could be written like so:

            flag Test : uint16 {
                A, B=4, C
            };
    """

    def __call__(self, value: Union[int, BinaryIO]) -> FlagInstance:
        if isinstance(value, int):
            return FlagInstance(self, value)

        return super().__call__(value)


class FlagInstance(EnumInstance):
    """Implements a value instance of a Flag"""

    def __bool__(self):
        return bool(self.value)

    __nonzero__ = __bool__

    def __or__(self, other: Union[int, FlagInstance]) -> FlagInstance:
        if hasattr(other, "value"):
            other = other.value

        return self.__class__(self.enum, self.value | other)

    def __and__(self, other: Union[int, FlagInstance]) -> FlagInstance:
        if hasattr(other, "value"):
            other = other.value

        return self.__class__(self.enum, self.value & other)

    def __xor__(self, other: Union[int, FlagInstance]) -> FlagInstance:
        if hasattr(other, "value"):
            other = other.value

        return self.__class__(self.enum, self.value ^ other)

    __ror__ = __or__
    __rand__ = __and__
    __rxor__ = __xor__

    def __invert__(self) -> FlagInstance:
        return self.__class__(self.enum, ~self.value)

    def __str__(self) -> str:
        if self.name is not None:
            return f"{self.enum.name}.{self.name}"

        members, _ = self.decompose()
        members_str = "|".join([str(name or value) for name, value in members])
        return f"{self.enum.name}.{members_str}"

    def __repr__(self) -> str:
        if self.name is not None:
            return f"<{self.enum.name}.{self.name}: {self.value}>"

        members, _ = self.decompose()
        members_str = "|".join([str(name or value) for name, value in members])
        return f"<{self.enum.name}.{members_str}: {self.value}>"

    @property
    def name(self) -> str:
        return self.enum.reverse.get(self.value, None)

    def decompose(self) -> Tuple[List[str], int]:
        members = []
        not_covered = self.value

        for name, value in self.enum.values.items():
            if value and ((value & self.value) == value):
                members.append((name, value))
                not_covered &= ~value

        if not members:
            members.append((None, self.value))

        members.sort(key=lambda m: m[0], reverse=True)
        if len(members) > 1 and members[0][1] == self.value:
            members.pop(0)

        return members, not_covered



import pytest
import ctypes as _ctypes

from dissect import cstruct


DUMMY_CSTRUCT = cstruct.cstruct()
PACKED_TYPE_INT8 = cstruct.PackedType(DUMMY_CSTRUCT, "int8", 1, "b")


@pytest.mark.parametrize(
    "cstruct_type, ctypes_type",
    [
        (PACKED_TYPE_INT8, _ctypes.c_int8),
        (cstruct.CharType(DUMMY_CSTRUCT), _ctypes.c_char),
        (cstruct.Array(DUMMY_CSTRUCT, PACKED_TYPE_INT8, 3), _ctypes.c_int8 * 3),
        (cstruct.Pointer(DUMMY_CSTRUCT, PACKED_TYPE_INT8), _ctypes.POINTER(_ctypes.c_int8)),
    ],
)
def test_ctypes_type(cstruct_type, ctypes_type):
    assert ctypes_type == cstruct.ctypes_type(cstruct_type)


def test_ctypes_type_exception():
    with pytest.raises(NotImplementedError):
        cstruct.ctypes_type("FAIL")



import sys

from dissect.cstruct import cstruct, dumpstruct

disk_def = """
#define MAX_MBR_CODE_SIZE 0x1b6
#define MBR_SIZE 0x200
#define PART_DOS_EXTD 5
#define PART_WIN_EXTD_LBA 0xF
#define PART_LINUX_EXTD 0x85

typedef struct sig_s {
    char        str[13];
} sig;

typedef struct part_s {
    uint8       bootable;               // +0: 0x80/0x00 - bootable/not bootable
    uint8   start_head;         // +1: head (start)
    uint16  start_cyl_sec;  // +2: cyl+sect (start)
    uint8   type;               // +4: type
    uint8   end_head;           // +5: head (end)
    uint16  end_cyl_sec;    // +6: cyl+sec (end)
    uint32  sector_ofs;         // +8: offset in sectors
    uint32  sector_size;    // +12: size in sectors
} part;

typedef struct mbr_s {
    uint8       jmp[3];
    sig     signature;
    uint8   crlf[3];            /* 10 */
    uint8   default_char;   /* 13 */
    uint8   chars[4];           /* 14 */
    uint16  delay;              /* 18 */
    uint16  offsets[4];         /* 1a..20 */
    char    rest_of_code[0x1b6-0x22];
    uint16  pad1;
    uint32  vol_no;
    uint16  pad2;
    part    part[4];
    uint16  bootsig;
} mbr;

// http://en.wikipedia.org/wiki/GUID_Partition_Table
struct GPT_HEADER {
    char        signature[8];
    uint32      revision;
    uint32      header_size;
    uint32      crc32;
    uint32      reserved;
    uint64      current_lba;
    uint64      backup_lba;
    uint64      first_usable_lba;
    uint64      last_usable_lba;
    uint8       guid[16];
    uint64      lba_partition_array;
    uint32      partition_table_count;
    uint32      partition_entry_size;
    uint32      partition_table_crc;
    char        _[420];
};

struct GPT_PARTITION {
    uint8       type_guid[16];
    char        partition_guid[16];
    uint64      first_lba;
    uint64      last_lba;
    uint64      attribute_flags;
    wchar       name[36];
};

// 0 (0x00)        16 bytes        Partition type GUID
// 16 (0x10)   16 bytes    Unique partition GUID
// 32 (0x20)   8 bytes     First LBA (little endian)
// 40 (0x28)   8 bytes     Last LBA (inclusive, usually odd)
// 48 (0x30)   8 bytes     Attribute flags (e.g. bit 60 denotes read-only)
// 56 (0x38)   72 bytes    Partition name (36 UTF-16LE code units)
"""

c_disk = cstruct()
c_disk.load(disk_def)

SECTOR_SIZE = 512


class Partition:
    def __init__(self, disk, offset, size, vtype, name, guid=None):
        self.disk = disk
        self.offset = offset
        self.size = size
        self.type = vtype
        self.name = name
        self.guid = guid

    def __repr__(self):
        return f"<Partition offset=0x{self.offset:x} size=0x{self.size:x} type={self.type} name={self.name}>"


def partitions(fh_part, mbr_part, offset):
    for mbr_p in mbr_part.part:
        part_offset = offset + mbr_p.sector_ofs * SECTOR_SIZE

        if mbr_p.type == 0x00:
            continue

        if mbr_p.type == 0x05:
            fh_part.seek(part_offset)
            e_mbr = c_disk.mbr(fh_part)
            for y_part in partitions(fh_part, e_mbr, part_offset):
                yield y_part

        yield Partition(fh_part, part_offset, mbr_p.sector_size * SECTOR_SIZE, mbr_p.type, None)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        sys.exit("usage: disk.py <disk or image>")

    fh = open(sys.argv[1], "rb")
    mbr = c_disk.mbr(fh)

    if mbr.bootsig != 0xAA55:
        sys.exit("Not a valid MBR")

    dumpstruct(mbr)

    for p in partitions(fh, mbr, 0):
        if p.type == 0xEE:
            fh.seek(p.offset)
            gpt = c_disk.GPT_HEADER(fh)
            dumpstruct(gpt)

            fh.seek(gpt.lba_partition_array * SECTOR_SIZE)
            for _ in range(gpt.partition_table_count):
                p = c_disk.GPT_PARTITION(fh)
                if p.first_lba == 0:
                    break

                part = Partition(
                    fh,
                    p.first_lba * SECTOR_SIZE,
                    (p.last_lba - p.first_lba) * SECTOR_SIZE,
                    p.type_guid,
                    p.name.rstrip("\x00"),
                    guid=p.partition_guid,
                )
                print(part)

            continue

        print(p)



from io import BytesIO
from typing import Any, BinaryIO, Dict

from dissect.cstruct.types import BaseType


class Instance:
    """Holds parsed structure data."""

    __slots__ = ("_type", "_values", "_sizes")

    def __init__(self, type_: BaseType, values: Dict[str, Any], sizes: Dict[str, int] = None):
        # Done in this manner to check if the attr is in the lookup
        object.__setattr__(self, "_type", type_)
        object.__setattr__(self, "_values", values)
        object.__setattr__(self, "_sizes", sizes)

    def __getattr__(self, attr: str) -> Any:
        try:
            return self._values[attr]
        except KeyError:
            raise AttributeError(f"Invalid attribute: {attr}")

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr not in self._type.lookup:
            raise AttributeError(f"Invalid attribute: {attr}")

        self._values[attr] = value

    def __getitem__(self, item: str) -> Any:
        return self._values[item]

    def __contains__(self, attr: str) -> bool:
        return attr in self._values

    def __repr__(self) -> str:
        values = ", ".join([f"{k}={hex(v) if isinstance(v, int) else repr(v)}" for k, v in self._values.items()])
        return f"<{self._type.name} {values}>"

    def __len__(self) -> int:
        return len(self.dumps())

    def __bytes__(self) -> bytes:
        return self.dumps()

    def _size(self, field: str) -> int:
        return self._sizes[field]

    def write(self, stream: BinaryIO) -> int:
        """Write this structure to a writable file-like object.

        Args:
            fh: File-like objects that supports writing.

        Returns:
            The amount of bytes written.
        """
        return self._type.write(stream, self)

    def dumps(self) -> bytes:
        """Dump this structure to a byte string.

        Returns:
            The raw bytes of this structure.
        """
        s = BytesIO()
        self.write(s)
        return s.getvalue()



from dissect import cstruct

from .utils import verify_compiled


def test_flag():
    cdef = """
    flag Test {
        a,
        b,
        c,
        d
    };

    flag Odd {
        a = 2,
        b,
        c,
        d = 32, e, f,
        g
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef)

    assert cs.Test.a == 1
    assert cs.Test.b == 2
    assert cs.Test.c == 4
    assert cs.Test.d == 8

    assert cs.Odd.a == 2
    assert cs.Odd.b == 4
    assert cs.Odd.c == 8
    assert cs.Odd.d == 32
    assert cs.Odd.e == 64
    assert cs.Odd.f == 128
    assert cs.Odd.g == 256

    assert cs.Test.a == cs.Test.a
    assert cs.Test.a != cs.Test.b
    assert bool(cs.Test(0)) is False
    assert bool(cs.Test(1)) is True

    assert cs.Test.a | cs.Test.b == 3
    assert str(cs.Test.c | cs.Test.d) == "Test.d|c"
    assert repr(cs.Test.a | cs.Test.b) == "<Test.b|a: 3>"
    assert cs.Test(2) == cs.Test.b
    assert cs.Test(3) == cs.Test.a | cs.Test.b
    assert cs.Test.c & 12 == cs.Test.c
    assert cs.Test.b & 12 == 0
    assert cs.Test.b ^ cs.Test.a == cs.Test.a | cs.Test.b

    assert ~cs.Test.a == -2
    assert str(~cs.Test.a) == "Test.d|c|b"


def test_flag_read(compiled):
    cdef = """
    flag Test16 : uint16 {
        A = 0x1,
        B = 0x2
    };

    flag Test24 : uint24 {
        A = 0x1,
        B = 0x2
    };

    flag Test32 : uint32 {
        A = 0x1,
        B = 0x2
    };

    struct test {
        Test16  a16;
        Test16  b16;
        Test24  a24;
        Test24  b24;
        Test32  a32;
        Test32  b32;
        Test16  l[2];
        Test16  c16;
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    buf = b"\x01\x00\x02\x00\x01\x00\x00\x02\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x01\x00\x02\x00\x03\x00"
    obj = cs.test(buf)

    assert obj.a16.enum == cs.Test16 and obj.a16.value == cs.Test16.A
    assert obj.b16.enum == cs.Test16 and obj.b16.value == cs.Test16.B
    assert obj.a24.enum == cs.Test24 and obj.a24.value == cs.Test24.A
    assert obj.b24.enum == cs.Test24 and obj.b24.value == cs.Test24.B
    assert obj.a32.enum == cs.Test32 and obj.a32.value == cs.Test32.A
    assert obj.b32.enum == cs.Test32 and obj.b32.value == cs.Test32.B

    assert len(obj.l) == 2
    assert obj.l[0].enum == cs.Test16 and obj.l[0].value == cs.Test16.A
    assert obj.l[1].enum == cs.Test16 and obj.l[1].value == cs.Test16.B

    assert obj.c16 == cs.Test16.A | cs.Test16.B
    assert obj.c16 & cs.Test16.A
    assert str(obj.c16) == "Test16.B|A"

    assert obj.dumps() == buf



from dissect.cstruct import utils


def test_pack_unpack():
    endian = "little"
    sign = False
    assert utils.p8(1, endian) == b"\x01"
    assert utils.p16(1, endian) == b"\x01\x00"
    assert utils.p32(1, endian) == b"\x01\x00\x00\x00"
    assert utils.p64(1, endian) == b"\x01\x00\x00\x00\x00\x00\x00\x00"
    assert utils.u8(b"\x01", endian, sign) == 1
    assert utils.u16(b"\x01\x00", endian, sign) == 1
    assert utils.u32(b"\x01\x00\x00\x00", endian, sign) == 1
    assert utils.u64(b"\x01\x00\x00\x00\x00\x00\x00\x00", endian, sign) == 1

    endian = "big"
    sign = False
    assert utils.p8(1, endian) == b"\x01"
    assert utils.p16(1, endian) == b"\x00\x01"
    assert utils.p32(1, endian) == b"\x00\x00\x00\x01"
    assert utils.p64(1, endian) == b"\x00\x00\x00\x00\x00\x00\x00\x01"
    assert utils.u8(b"\x01", endian, sign) == 1
    assert utils.u16(b"\x00\x01", endian, sign) == 1
    assert utils.u32(b"\x00\x00\x00\x01", endian, sign) == 1
    assert utils.u64(b"\x00\x00\x00\x00\x00\x00\x00\x01", endian, sign) == 1

    endian = "network"
    sign = False
    assert utils.p8(1, endian) == b"\x01"
    assert utils.p16(1, endian) == b"\x00\x01"
    assert utils.p32(1, endian) == b"\x00\x00\x00\x01"
    assert utils.p64(1, endian) == b"\x00\x00\x00\x00\x00\x00\x00\x01"
    assert utils.u8(b"\x01", endian, sign) == 1
    assert utils.u16(b"\x00\x01", endian, sign) == 1
    assert utils.u32(b"\x00\x00\x00\x01", endian, sign) == 1
    assert utils.u64(b"\x00\x00\x00\x00\x00\x00\x00\x01", endian, sign) == 1

    endian = "little"
    sign = True
    assert utils.p8(-120, endian) == b"\x88"
    assert utils.p16(-120, endian) == b"\x88\xff"
    assert utils.p32(-120, endian) == b"\x88\xff\xff\xff"
    assert utils.p64(-120, endian) == b"\x88\xff\xff\xff\xff\xff\xff\xff"
    assert utils.u8(b"\x88", endian, sign) == -120
    assert utils.u16(b"\x88\xff", endian, sign) == -120
    assert utils.u32(b"\x88\xff\xff\xff", endian, sign) == -120
    assert utils.u64(b"\x88\xff\xff\xff\xff\xff\xff\xff", endian, sign) == -120

    endian = "big"
    sign = True
    assert utils.p8(-120, endian) == b"\x88"
    assert utils.p16(-120, endian) == b"\xff\x88"
    assert utils.p32(-120, endian) == b"\xff\xff\xff\x88"
    assert utils.p64(-120, endian) == b"\xff\xff\xff\xff\xff\xff\xff\x88"
    assert utils.u8(b"\x88", endian, sign) == -120
    assert utils.u16(b"\xff\x88", endian, sign) == -120
    assert utils.u32(b"\xff\xff\xff\x88", endian, sign) == -120
    assert utils.u64(b"\xff\xff\xff\xff\xff\xff\xff\x88", endian, sign) == -120

    assert utils.pack(1, 24) == b"\x01\x00\x00"
    assert utils.unpack(b"\x01\x00\x00", 24) == 1

    assert utils.pack(213928798) == b"^K\xc0\x0c"
    assert utils.unpack(b"^K\xc0\x0c") == 213928798


def test_swap():
    assert utils.swap16(0x0001) == 0x0100
    assert utils.swap32(0x00000001) == 0x01000000
    assert utils.swap64(0x0000000000000001) == 0x0100000000000000



from io import BytesIO

import pytest

from dissect import cstruct

from .utils import verify_compiled


def test_struct_simple(compiled):
    cdef = """
    struct test {
        char    magic[4];
        wchar   wmagic[4];
        uint8   a;
        uint16  b;
        uint32  c;
        char    string[];
        wchar   wstring[];
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    buf = b"testt\x00e\x00s\x00t\x00\x01\x02\x03\x04\x05\x06\x07lalala\x00t\x00e\x00s\x00t\x00\x00\x00"
    obj = cs.test(buf)

    assert "magic" in obj
    assert obj.magic == b"test"
    assert obj["magic"] == obj.magic
    assert obj.wmagic == "test"
    assert obj.a == 0x01
    assert obj.b == 0x0302
    assert obj.c == 0x07060504
    assert obj.string == b"lalala"
    assert obj.wstring == "test"

    with pytest.raises(AttributeError):
        obj.nope

    assert obj._size("magic") == 4
    assert len(obj) == len(buf)
    assert obj.dumps() == buf

    assert repr(obj)

    fh = BytesIO()
    obj.write(fh)
    assert fh.getvalue() == buf


def test_struct_simple_be(compiled):
    cdef = """
    struct test {
        char    magic[4];
        wchar   wmagic[4];
        uint8   a;
        uint16  b;
        uint32  c;
        char    string[];
        wchar   wstring[];
    };
    """
    cs = cstruct.cstruct(endian=">")
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    buf = b"test\x00t\x00e\x00s\x00t\x01\x02\x03\x04\x05\x06\x07lalala\x00\x00t\x00e\x00s\x00t\x00\x00"
    obj = cs.test(buf)

    assert obj.magic == b"test"
    assert obj.wmagic == "test"
    assert obj.a == 0x01
    assert obj.b == 0x0203
    assert obj.c == 0x04050607
    assert obj.string == b"lalala"
    assert obj.wstring == "test"
    assert obj.dumps() == buf


def test_struct_definitions(compiled):
    cdef = """
    struct _test {
        uint32  a;
        // uint32 comment
        uint32  b;
    } test, test1;
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    assert cs._test == cs.test == cs.test1
    assert cs.test.name == "_test"
    assert cs._test.name == "_test"

    assert "a" in cs.test.lookup
    assert "b" in cs.test.lookup

    with pytest.raises(cstruct.ParserError):
        cdef = """
        struct {
            uint32  a;
        };
        """
        cs.load(cdef)


def test_struct_expressions(compiled):
    cdef = """
    #define const 1
    struct test {
        uint8   flag;
        uint8   data_1[flag & 1 * 4];
        uint8   data_2[flag & (1 << 2)];
        uint8   data_3[const];
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    obj = cs.test(b"\x01\x00\x01\x02\x03\xff")
    assert obj.flag == 1
    assert obj.data_1 == [0, 1, 2, 3]
    assert obj.data_2 == []
    assert obj.data_3 == [255]

    obj = cs.test(b"\x04\x04\x05\x06\x07\xff")
    assert obj.flag == 4
    assert obj.data_1 == []
    assert obj.data_2 == [4, 5, 6, 7]
    assert obj.data_3 == [255]


def test_struct_sizes(compiled):
    cdef = """
    struct static {
        uint32  test;
    };

    struct dynamic {
        uint32  test[];
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.static, compiled)
    assert verify_compiled(cs.dynamic, compiled)

    assert len(cs.static) == 4

    if not compiled:
        cs.static.add_field("another", cs.uint32)
        assert len(cs.static) == 8
        cs.static.add_field("atoffset", cs.uint32, offset=12)
        assert len(cs.static) == 16

        obj = cs.static(b"\x01\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x03\x00\x00\x00")
        assert obj.test == 1
        assert obj.another == 2
        assert obj.atoffset == 3

        with pytest.raises(TypeError) as excinfo:
            len(cs.dynamic)
        assert str(excinfo.value) == "Dynamic size"
    else:
        with pytest.raises(NotImplementedError) as excinfo:
            cs.static.add_field("another", cs.uint32)
        assert str(excinfo.value) == "Can't add fields to a compiled structure"


def test_struct_nested(compiled):
    cdef = """
    struct test_named {
        char magic[4];
        struct {
            uint32 a;
            uint32 b;
        } a;
        struct {
            char   c[8];
        } b;
    };

    struct test_anonymous {
        char magic[4];
        struct {
            uint32 a;
            uint32 b;
        };
        struct {
            char   c[8];
        };
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test_named, compiled)
    assert verify_compiled(cs.test_anonymous, compiled)

    assert len(cs.test_named) == len(cs.test_anonymous) == 20

    data = b"zomg\x39\x05\x00\x00\x28\x23\x00\x00deadbeef"
    obj = cs.test_named(data)
    assert obj.magic == b"zomg"
    assert obj.a.a == 1337
    assert obj.a.b == 9000
    assert obj.b.c == b"deadbeef"
    assert obj.dumps() == data

    obj = cs.test_anonymous(data)
    assert obj.magic == b"zomg"
    assert obj.a == 1337
    assert obj.b == 9000
    assert obj.c == b"deadbeef"
    assert obj.dumps() == data


def test_struct_write(compiled):
    cdef = """
    struct test {
        char    magic[4];
        wchar   wmagic[4];
        uint8   a;
        uint16  b;
        uint32  c;
        char    string[];
        wchar   wstring[];
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    buf = b"testt\x00e\x00s\x00t\x00\x01\x02\x03\x04\x05\x06\x07lalala\x00t\x00e\x00s\x00t\x00\x00\x00"

    obj = cs.test()
    obj.magic = "test"
    obj.wmagic = "test"
    obj.a = 0x01
    obj.b = 0x0302
    obj.c = 0x07060504
    obj.string = b"lalala"
    obj.wstring = "test"

    with pytest.raises(AttributeError):
        obj.nope = 1

    assert obj.dumps() == buf

    inst = cs.test(
        magic=b"test",
        wmagic="test",
        a=0x01,
        b=0x0302,
        c=0x07060504,
        string=b"lalala",
        wstring="test",
    )
    assert inst.dumps() == buf


def test_struct_write_be(compiled):
    cdef = """
    struct test {
        char    magic[4];
        wchar   wmagic[4];
        uint8   a;
        uint16  b;
        uint32  c;
        char    string[];
        wchar   wstring[];
    };
    """
    cs = cstruct.cstruct(endian=">")
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    buf = b"test\x00t\x00e\x00s\x00t\x01\x02\x03\x04\x05\x06\x07lalala\x00\x00t\x00e\x00s\x00t\x00\x00"

    obj = cs.test()
    obj.magic = "test"
    obj.wmagic = "test"
    obj.a = 0x01
    obj.b = 0x0203
    obj.c = 0x04050607
    obj.string = b"lalala"
    obj.wstring = "test"

    assert obj.dumps() == buf


def test_struct_write_anonymous():
    cdef = """
    struct test {
        uint32 a;
        union {
            struct {
                uint16 b1;
                uint16 b2;
            };
            uint32 b;
        };
        uint32 c;
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef)

    obj = cs.test(a=1, c=3)
    assert obj.dumps() == b"\x01\x00\x00\x00\x00\x00\x00\x00\x03\x00\x00\x00"



from __future__ import annotations

from typing import TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from dissect.cstruct import cstruct


class Expression:
    """Expression parser for simple calculations in definitions."""

    operators = [
        ("*", lambda a, b: a * b),
        ("/", lambda a, b: a // b),
        ("%", lambda a, b: a % b),
        ("+", lambda a, b: a + b),
        ("-", lambda a, b: a - b),
        (">>", lambda a, b: a >> b),
        ("<<", lambda a, b: a << b),
        ("&", lambda a, b: a & b),
        ("^", lambda a, b: a ^ b),
        ("|", lambda a, b: a | b),
    ]

    def __init__(self, cstruct: cstruct, expression: str):
        self.cstruct = cstruct
        self.expression = expression

    def __repr__(self) -> str:
        return self.expression

    def evaluate(self, context: Dict[str, int] = None) -> int:
        context = context or {}
        levels = []
        buf = ""

        for i in range(len(self.expression)):
            if self.expression[i] == "(":
                levels.append(buf)
                buf = ""
                continue

            if self.expression[i] == ")":
                if levels[-1] == "sizeof":
                    value = len(self.cstruct.resolve(buf))
                    levels[-1] = ""
                else:
                    value = self.evaluate_part(buf, context)
                buf = levels.pop()
                buf += str(value)
                continue

            buf += self.expression[i]

        return self.evaluate_part(buf, context)

    def evaluate_part(self, buf: str, context: Dict[str, int]) -> int:
        buf = buf.strip()

        # Very simple way to support an expression(part) that is a single,
        # negative value. To use negative values in more complex expressions,
        # they must be wrapped in brackets, e.g.: 2 * (-5).
        #
        # To have full support for the negation operator a proper expression
        # parser must be build.
        if buf.startswith("-") and buf[1:].isnumeric():
            return int(buf)

        for operator in self.operators:
            if operator[0] in buf:
                a, b = buf.rsplit(operator[0], 1)

                return operator[1](self.evaluate_part(a, context), self.evaluate_part(b, context))

        if buf in context:
            return context[buf]

        if buf.startswith("0x"):
            return int(buf, 16)

        if buf in self.cstruct.consts:
            return self.cstruct.consts[buf]

        return int(buf)



from __future__ import annotations

import io
from collections import OrderedDict
from typing import BinaryIO, List, TYPE_CHECKING

from dissect.cstruct.bitbuffer import BitBuffer
from dissect.cstruct.types import Array, BaseType, Enum, Instance, Pointer

if TYPE_CHECKING:
    from dissect.cstruct import cstruct


class Field:
    """Holds a structure field."""

    def __init__(self, name: str, type_: BaseType, bits: int = None, offset: int = None):
        self.name = name
        self.type = type_
        self.bits = bits
        self.offset = offset
        self.alignment = type_.alignment

    def __repr__(self):
        bits_str = " : {self.bits}" if self.bits else ""
        return f"<Field {self.name} {self.type}{bits_str}>"


class Structure(BaseType):
    """Type class for structures."""

    def __init__(
        self, cstruct: cstruct, name: str, fields: List[Field] = None, align: bool = False, anonymous: bool = False
    ):
        super().__init__(cstruct)
        self.name = name
        self.size = None
        self.alignment = None

        self.lookup = OrderedDict()
        self.fields = fields

        self.align = align
        self.anonymous = anonymous
        self.dynamic = False

        for field in self.fields:
            self.lookup[field.name] = field
            if isinstance(field.type, Structure) and field.type.anonymous:
                self.lookup.update(field.type.lookup)

        self._calc_size_and_offsets()

    def __len__(self) -> int:
        if self.dynamic:
            raise TypeError("Dynamic size")

        if self.size is None:
            self._calc_size_and_offsets()

        return self.size

    def __repr__(self) -> str:
        return f"<Structure {self.name}>"

    def _calc_size_and_offsets(self) -> None:
        """Iterate all fields in this structure to calculate the field offsets and total structure size.

        If a structure has a dynamic field, further field offsets will be set to None and self.dynamic
        will be set to True.
        """
        # The current offset, set to None if we become dynamic
        offset = 0
        # The current alignment for this structure
        alignment = 0

        # The current bit field type
        bits_type = None
        # The offset of the current bit field, set to None if we become dynamic
        bits_field_offset = 0
        # How many bits we have left in the current bit field
        bits_remaining = 0

        for field in self.fields:
            if field.offset is not None:
                # If a field already has an offset, it's leading
                offset = field.offset

            if self.align and offset is not None:
                # Round to next alignment
                offset += -offset & (field.alignment - 1)

            # The alignment of this struct is equal to its largest members' alignment
            alignment = max(alignment, field.type.alignment)

            if field.bits:
                field_type = field.type

                if isinstance(field_type, Enum):
                    field_type = field_type.type

                # Bit fields have special logic
                if (
                    # Exhausted a bit field
                    bits_remaining == 0
                    # Moved to a bit field of another type, e.g. uint16 f1 : 8, uint32 f2 : 8;
                    or field_type != bits_type
                    # Still processing a bit field, but it's at a different offset due to alignment or a manual offset
                    or (bits_type is not None and offset > bits_field_offset + bits_type.size)
                ):
                    # ... if any of this is true, we have to move to the next field
                    bits_type = field_type
                    bits_count = bits_type.size * 8
                    bits_remaining = bits_count
                    bits_field_offset = offset

                    if offset is not None:
                        # We're not dynamic, update the structure size and current offset
                        offset += bits_type.size

                    field.offset = bits_field_offset

                bits_remaining -= field.bits

                if bits_remaining < 0:
                    raise ValueError("Straddled bit fields are unsupported")
            else:
                # Reset bits stuff
                bits_type = None
                bits_field_offset = bits_remaining = 0

                field.offset = offset

                if offset is not None:
                    # We're not dynamic, update the structure size and current offset
                    try:
                        field_len = len(field.type)
                    except TypeError:
                        # This field is dynamic
                        offset = None
                        self.dynamic = True
                        continue

                    offset += field_len

        if self.align and offset is not None:
            # Add "tail padding" if we need to align
            # This bit magic rounds up to the next alignment boundary
            # E.g. offset = 3; alignment = 8; -offset & (alignment - 1) = 5
            offset += -offset & (alignment - 1)

        # The structure size is whatever the currently calculated offset is
        self.size = offset
        self.alignment = alignment

    def _read(self, stream: BinaryIO, *args, **kwargs) -> Instance:
        bit_buffer = BitBuffer(stream, self.cstruct.endian)
        struct_start = stream.tell()

        result = OrderedDict()
        sizes = {}
        for field in self.fields:
            offset = stream.tell()
            field_type = self.cstruct.resolve(field.type)

            if field.offset and offset != struct_start + field.offset:
                # Field is at a specific offset, either alligned or added that way
                offset = struct_start + field.offset
                stream.seek(offset)

            if self.align and field.offset is None:
                # Previous field was dynamically sized and we need to align
                offset += -offset & (field.alignment - 1)
                stream.seek(offset)

            if field.bits:
                if isinstance(field_type, Enum):
                    value = field_type(bit_buffer.read(field_type.type, field.bits))
                else:
                    value = bit_buffer.read(field_type, field.bits)

                if field.name:
                    result[field.name] = value
                continue
            else:
                bit_buffer.reset()

            if isinstance(field_type, (Array, Pointer)):
                value = field_type._read(stream, result)
            else:
                value = field_type._read(stream)

            if isinstance(field_type, Structure) and field_type.anonymous:
                sizes.update(value._sizes)
                result.update(value._values)
            else:
                if field.name:
                    sizes[field.name] = stream.tell() - offset
                    result[field.name] = value

        if self.align:
            # Align the stream
            stream.seek(-stream.tell() & (self.alignment - 1), io.SEEK_CUR)

        return Instance(self, result, sizes)

    def _write(self, stream: BinaryIO, data: Instance) -> int:
        bit_buffer = BitBuffer(stream, self.cstruct.endian)
        struct_start = stream.tell()
        num = 0

        for field in self.fields:
            bit_field_type = (field.type.type if isinstance(field.type, Enum) else field.type) if field.bits else None
            # Current field is not a bit field, but previous was
            # Or, moved to a bit field of another type, e.g. uint16 f1 : 8, uint32 f2 : 8;
            if (not field.bits and bit_buffer._type is not None) or (
                bit_buffer._type and bit_buffer._type != bit_field_type
            ):
                # Flush the current bit buffer so we can process alignment properly
                bit_buffer.flush()

            offset = stream.tell()

            if field.offset and offset < struct_start + field.offset:
                # Field is at a specific offset, either alligned or added that way
                stream.write(b"\x00" * (struct_start + field.offset - offset))
                offset = struct_start + field.offset

            if self.align and field.offset is None:
                is_bitbuffer_boundary = bit_buffer._type and (
                    bit_buffer._remaining == 0 or bit_buffer._type != field.type
                )
                if not bit_buffer._type or is_bitbuffer_boundary:
                    # Previous field was dynamically sized and we need to align
                    align_pad = -offset & (field.alignment - 1)
                    stream.write(b"\x00" * align_pad)
                    offset += align_pad

            value = getattr(data, field.name, None)
            if value is None:
                value = field.type.default()

            if field.bits:
                if isinstance(field.type, Enum):
                    bit_buffer.write(field.type.type, value.value, field.bits)
                else:
                    bit_buffer.write(field.type, value, field.bits)
            else:
                if isinstance(field.type, Structure) and field.type.anonymous:
                    field.type._write(stream, data)
                else:
                    field.type._write(stream, value)
                num += stream.tell() - offset

        if bit_buffer._type is not None:
            bit_buffer.flush()

        if self.align:
            # Align the stream
            stream.write(b"\x00" * (-stream.tell() & (self.alignment - 1)))

        return num

    def add_field(self, name: str, type_: BaseType, offset: int = None) -> None:
        """Add a field to this structure.

        Args:
            name: The field name.
            type_: The field type.
            offset: The field offset.
        """
        field = Field(name, type_, offset=offset)
        self.fields.append(field)
        self.lookup[name] = field
        if isinstance(field.type, Structure) and field.type.anonymous:
            self.lookup.update(field.type.lookup)
        self.size = None

    def default(self) -> Instance:
        """Create and return an empty Instance from this structure.

        Returns:
            An empty Instance from this structure.
        """
        result = OrderedDict()
        for field in self.fields:
            if isinstance(field.type, Structure) and field.type.anonymous:
                result.update(field.type.default()._values)
            else:
                result[field.name] = field.type.default()

        return Instance(self, result)

    def show(self, indent: int = 0) -> None:
        """Pretty print this structure."""
        if indent == 0:
            print(f"struct {self.name}")

        for field in self.fields:
            if field.offset is None:
                offset = "0x??"
            else:
                offset = f"0x{field.offset:02x}"

            print(f"{' ' * indent}+{offset} {field.name} {field.type}")

            if isinstance(field.type, Structure):
                field.type.show(indent + 1)


class Union(Structure):
    """Type class for unions"""

    def __repr__(self) -> str:
        return f"<Union {self.name}>"

    def _calc_size_and_offsets(self) -> None:
        self.size = max(len(field.type) for field in self.fields)
        self.alignment = max(field.type.alignment for field in self.fields)

    def _read(self, stream: BinaryIO) -> Instance:
        buf = io.BytesIO(memoryview(stream.read(len(self))))
        result = OrderedDict()
        sizes = {}

        for field in self.fields:
            start = 0
            buf.seek(0)
            field_type = self.cstruct.resolve(field.type)

            if field.offset:
                buf.seek(field.offset)
                start = field.offset

            if isinstance(field_type, (Array, Pointer)):
                v = field_type._read(buf, result)
            else:
                v = field_type._read(buf)

            if isinstance(field_type, Structure) and field_type.anonymous:
                sizes.update(v._sizes)
                result.update(v._values)
            else:
                sizes[field.name] = buf.tell() - start
                result[field.name] = v

        return Instance(self, result, sizes)

    def _write(self, stream: BinaryIO, data: Instance) -> Instance:
        offset = stream.tell()

        # Find the largest field
        field = max(self.fields, key=lambda e: len(e.type))

        # Write the value to the stream using the largest field type
        if isinstance(field.type, Structure) and field.type.anonymous:
            field.type._write(stream, data)
        else:
            field.type._write(stream, getattr(data, field.name))

        return stream.tell() - offset

    def show(self, indent: int = 0) -> None:
        raise NotImplementedError()



import pytest

from dissect import cstruct
from dissect.cstruct.types.pointer import PointerInstance

from .utils import verify_compiled


@pytest.mark.parametrize("compiled", [True, False])
def test_pointer_basic(compiled):
    d = """
    struct ptrtest {
        uint32  *ptr1;
        uint32  *ptr2;
    };
    """
    c = cstruct.cstruct(pointer="uint16")
    c.load(d, compiled=compiled)

    assert verify_compiled(c.ptrtest, compiled)

    d = b"\x04\x00\x08\x00\x01\x02\x03\x04\x05\x06\x07\x08"
    p = c.ptrtest(d)

    assert p.ptr1 != 0
    assert p.ptr2 != 0
    assert p.ptr1 != p.ptr2
    assert p.ptr1 == 4
    assert p.ptr2 == 8
    assert p.ptr1.dereference() == 0x04030201
    assert p.ptr2.dereference() == 0x08070605

    p.ptr1 += 2
    p.ptr2 -= 2
    assert p.ptr1 == p.ptr2
    assert p.ptr1.dereference() == p.ptr2.dereference() == 0x06050403

    assert p.dumps() == b"\x06\x00\x06\x00"

    with pytest.raises(cstruct.NullPointerDereference):
        c.ptrtest(b"\x00\x00\x00\x00").ptr1.dereference()


@pytest.mark.parametrize("compiled", [True, False])
def test_pointer_struct(compiled):
    d = """
    struct test {
        char    magic[4];
        wchar   wmagic[4];
        uint8   a;
        uint16  b;
        uint32  c;
        char    string[];
        wchar   wstring[];
    };

    struct ptrtest {
        test    *ptr;
    };
    """
    c = cstruct.cstruct(pointer="uint16")
    c.load(d, compiled=compiled)

    assert verify_compiled(c.test, compiled)
    assert verify_compiled(c.ptrtest, compiled)

    d = b"\x02\x00testt\x00e\x00s\x00t\x00\x01\x02\x03\x04\x05\x06\x07lalala\x00t\x00e\x00s\x00t\x00\x00\x00"
    p = c.ptrtest(d)

    assert p.ptr != 0

    assert p.ptr.magic == b"test"
    assert p.ptr.wmagic == "test"
    assert p.ptr.a == 0x01
    assert p.ptr.b == 0x0302
    assert p.ptr.c == 0x07060504
    assert p.ptr.string == b"lalala"
    assert p.ptr.wstring == "test"

    assert p.dumps() == b"\x02\x00"

    with pytest.raises(cstruct.NullPointerDereference):
        c.ptrtest(b"\x00\x00\x00\x00").ptr.magic


def test_pointer_arithmetic():
    inst = PointerInstance(None, None, 0, None)
    assert inst._addr == 0

    inst += 4
    assert inst._addr == 4

    inst -= 2
    assert inst._addr == 2

    inst *= 12
    assert inst._addr == 24

    inst //= 2
    assert inst._addr == 12

    inst %= 10
    assert inst._addr == 2

    inst **= 4
    assert inst._addr == 16

    inst <<= 1
    assert inst._addr == 32

    inst >>= 2
    assert inst._addr == 8

    inst &= 2
    assert inst._addr == 0

    inst ^= 4
    assert inst._addr == 4

    inst |= 8
    assert inst._addr == 12



from dissect.cstruct.compiler import Compiler
from dissect.cstruct.expression import Expression
from dissect.cstruct.types.base import Array, BaseType, RawType
from dissect.cstruct.types.chartype import CharType
from dissect.cstruct.types.instance import Instance
from dissect.cstruct.types.structure import Structure, Field, Union
from dissect.cstruct.types.voidtype import VoidType
from dissect.cstruct.types.wchartype import WcharType
from dissect.cstruct.types.packedtype import PackedType
from dissect.cstruct.types.flag import Flag, FlagInstance
from dissect.cstruct.types.enum import Enum, EnumInstance
from dissect.cstruct.types.bytesinteger import BytesInteger
from dissect.cstruct.types.pointer import Pointer, PointerInstance

from dissect.cstruct.exceptions import (
    Error,
    ParserError,
    ResolveError,
    NullPointerDereference,
)

from dissect.cstruct.cstruct import (
    cstruct,
    ctypes,
    ctypes_type,
)

from dissect.cstruct.utils import (
    dumpstruct,
    hexdump,
    pack,
    p8,
    p16,
    p32,
    p64,
    swap,
    swap16,
    swap32,
    swap64,
    unpack,
    u8,
    u16,
    u32,
    u64,
)

from dissect.cstruct.bitbuffer import BitBuffer

__all__ = [
    "Compiler",
    "Array",
    "Union",
    "Field",
    "Instance",
    "Structure",
    "Expression",
    "PackedType",
    "Pointer",
    "PointerInstance",
    "VoidType",
    "WcharType",
    "RawType",
    "BaseType",
    "CharType",
    "Enum",
    "EnumInstance",
    "Flag",
    "FlagInstance",
    "BytesInteger",
    "BitBuffer",
    "cstruct",
    "ctypes",
    "ctypes_type",
    "dumpstruct",
    "hexdump",
    "pack",
    "p8",
    "p16",
    "p32",
    "p64",
    "swap",
    "swap16",
    "swap32",
    "swap64",
    "unpack",
    "u8",
    "u16",
    "u32",
    "u64",
    "Error",
    "ParserError",
    "ResolveError",
    "NullPointerDereference",
]



#!/usr/bin/env python
import sys

from dissect.cstruct import cstruct, dumpstruct

PE_DEF = """
#define IMAGE_NUMBEROF_DIRECTORY_ENTRIES 16
#define IMAGE_SIZEOF_SHORT_NAME          8

typedef struct _IMAGE_DOS_HEADER
{
    WORD e_magic;
    WORD e_cblp;
    WORD e_cp;
    WORD e_crlc;
    WORD e_cparhdr;
    WORD e_minalloc;
    WORD e_maxalloc;
    WORD e_ss;
    WORD e_sp;
    WORD e_csum;
    WORD e_ip;
    WORD e_cs;
    WORD e_lfarlc;
    WORD e_ovno;
    WORD e_res[4];
    WORD e_oemid;
    WORD e_oeminfo;
    WORD e_res2[10];
    LONG e_lfanew;
} IMAGE_DOS_HEADER;

typedef struct _IMAGE_FILE_HEADER {
    WORD  Machine;
    WORD  NumberOfSections;
    DWORD TimeDateStamp;
    DWORD PointerToSymbolTable;
    DWORD NumberOfSymbols;
    WORD  SizeOfOptionalHeader;
    WORD  Characteristics;
} IMAGE_FILE_HEADER;

typedef struct _IMAGE_DATA_DIRECTORY {
    ULONG   VirtualAddress;
    ULONG   Size;
} IMAGE_DATA_DIRECTORY;

typedef struct _IMAGE_OPTIONAL_HEADER {
    WORD                 Magic;
    BYTE                 MajorLinkerVersion;
    BYTE                 MinorLinkerVersion;
    DWORD                SizeOfCode;
    DWORD                SizeOfInitializedData;
    DWORD                SizeOfUninitializedData;
    DWORD                AddressOfEntryPoint;
    DWORD                BaseOfCode;
    DWORD                BaseOfData;
    DWORD                ImageBase;
    DWORD                SectionAlignment;
    DWORD                FileAlignment;
    WORD                 MajorOperatingSystemVersion;
    WORD                 MinorOperatingSystemVersion;
    WORD                 MajorImageVersion;
    WORD                 MinorImageVersion;
    WORD                 MajorSubsystemVersion;
    WORD                 MinorSubsystemVersion;
    DWORD                Win32VersionValue;
    DWORD                SizeOfImage;
    DWORD                SizeOfHeaders;
    DWORD                CheckSum;
    WORD                 Subsystem;
    WORD                 DllCharacteristics;
    DWORD                SizeOfStackReserve;
    DWORD                SizeOfStackCommit;
    DWORD                SizeOfHeapReserve;
    DWORD                SizeOfHeapCommit;
    DWORD                LoaderFlags;
    DWORD                NumberOfRvaAndSizes;
    IMAGE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
} IMAGE_OPTIONAL_HEADER;

typedef struct _IMAGE_OPTIONAL_HEADER64 {
    WORD        Magic;
    BYTE        MajorLinkerVersion;
    BYTE        MinorLinkerVersion;
    DWORD       SizeOfCode;
    DWORD       SizeOfInitializedData;
    DWORD       SizeOfUninitializedData;
    DWORD       AddressOfEntryPoint;
    DWORD       BaseOfCode;
    ULONGLONG   ImageBase;
    DWORD       SectionAlignment;
    DWORD       FileAlignment;
    WORD        MajorOperatingSystemVersion;
    WORD        MinorOperatingSystemVersion;
    WORD        MajorImageVersion;
    WORD        MinorImageVersion;
    WORD        MajorSubsystemVersion;
    WORD        MinorSubsystemVersion;
    DWORD       Win32VersionValue;
    DWORD       SizeOfImage;
    DWORD       SizeOfHeaders;
    DWORD       CheckSum;
    WORD        Subsystem;
    WORD        DllCharacteristics;
    ULONGLONG   SizeOfStackReserve;
    ULONGLONG   SizeOfStackCommit;
    ULONGLONG   SizeOfHeapReserve;
    ULONGLONG   SizeOfHeapCommit;
    DWORD       LoaderFlags;
    DWORD       NumberOfRvaAndSizes;
    IMAGE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
} IMAGE_OPTIONAL_HEADER64;

typedef struct _IMAGE_SECTION_HEADER {
    char    Name[IMAGE_SIZEOF_SHORT_NAME];
    ULONG   VirtualSize;
    ULONG   VirtualAddress;
    ULONG   SizeOfRawData;
    ULONG   PointerToRawData;
    ULONG   PointerToRelocations;
    ULONG   PointerToLinenumbers;
    USHORT  NumberOfRelocations;
    USHORT  NumberOfLinenumbers;
    ULONG   Characteristics;
} IMAGE_SECTION_HEADER;
"""

pestruct = cstruct()
pestruct.load(PE_DEF)


if __name__ == '__main__':
    if len(sys.argv) != 2:
        sys.exit("usage: pe.py <pe file>")

    fh = open(sys.argv[1], 'rb')

    mz = pestruct.IMAGE_DOS_HEADER(fh)
    if mz.e_magic != 0x5a4d:
        sys.exit("Invalid PE")

    fh.seek(mz.e_lfanew)
    signature = pestruct.uint32(fh)
    if signature != 0x4550:
        sys.exit("Invalid PE")

    file_header = pestruct.IMAGE_FILE_HEADER(fh)
    if file_header.Machine == 0x8664:
        optional_header = pestruct.IMAGE_OPTIONAL_HEADER64(fh)
    else:
        optional_header = pestruct.IMAGE_OPTIONAL_HEADER(fh)

    dumpstruct(mz)
    dumpstruct(file_header)
    dumpstruct(optional_header)

    sections = [pestruct.IMAGE_SECTION_HEADER(fh) for _ in range(file_header.NumberOfSections)]
    for s in sections:
        dumpstruct(s)



from __future__ import annotations

from typing import BinaryIO, List, TYPE_CHECKING

from dissect.cstruct.types import RawType

if TYPE_CHECKING:
    from dissect.cstruct import cstruct


class BytesInteger(RawType):
    """Implements an integer type that can span an arbitrary amount of bytes."""

    def __init__(self, cstruct: cstruct, name: str, size: int, signed: bool, alignment: int = None):
        self.signed = signed
        super().__init__(cstruct, name, size, alignment)

    @staticmethod
    def parse(buf: BinaryIO, size: int, count: int, signed: bool, endian: str) -> List[int]:
        nums = []

        for c in range(count):
            num = 0
            data = buf[c * size : (c + 1) * size]
            if endian == "<":  # little-endian (LE)
                data = b"".join(data[i : i + 1] for i in reversed(range(len(data))))

            ints = list(data)
            for i in ints:
                num = (num << 8) | i

            if signed and (num & (1 << (size * 8 - 1))):
                bias = 1 << (size * 8 - 1)
                num -= bias * 2

            nums.append(num)

        return nums

    @staticmethod
    def pack(data: List[int], size: int, endian: str, signed: bool) -> bytes:
        buf = []

        bits = size * 8
        unsigned_min = 0
        unsigned_max = (2**bits) - 1
        signed_min = -(2 ** (bits - 1))
        signed_max = (2 ** (bits - 1)) - 1

        for i in data:

            if signed and (i < signed_min or i > signed_max):
                raise OverflowError(f"{i} exceeds bounds for signed {bits} bits BytesInteger")
            elif not signed and (i < unsigned_min or i > unsigned_max):
                raise OverflowError(f"{i} exceeds bounds for unsigned {bits} bits BytesInteger")

            num = int(i)
            if num < 0:
                num += 1 << (size * 8)

            d = [b"\x00"] * size
            i = size - 1

            while i >= 0:
                b = num & 255
                d[i] = bytes((b,))
                num >>= 8
                i -= 1

            if endian == "<":
                d = b"".join(d[i : i + 1][0] for i in reversed(range(len(d))))
            else:
                d = b"".join(d)

            buf.append(d)

        return b"".join(buf)

    def _read(self, stream: BinaryIO) -> int:
        return self.parse(stream.read(self.size * 1), self.size, 1, self.signed, self.cstruct.endian)[0]

    def _read_array(self, stream: BinaryIO, count: int) -> List[int]:
        return self.parse(
            stream.read(self.size * count),
            self.size,
            count,
            self.signed,
            self.cstruct.endian,
        )

    def _read_0(self, stream: BinaryIO) -> List[int]:
        result = []

        while True:
            v = self._read(stream)
            if v == 0:
                break

            result.append(v)

        return result

    def _write(self, stream: BinaryIO, data: int) -> int:
        return stream.write(self.pack([data], self.size, self.cstruct.endian, self.signed))

    def _write_array(self, stream: BinaryIO, data: List[int]) -> int:
        return stream.write(self.pack(data, self.size, self.cstruct.endian, self.signed))

    def _write_0(self, stream: BinaryIO, data: List[int]) -> int:
        return self._write_array(stream, data + [0])

    def default(self) -> int:
        return 0

    def default_array(self, count: int) -> List[int]:
        return [0] * count



#!/usr/bin/env python
from dissect.cstruct import cstruct, dumpstruct

import socket
import struct

protocol = cstruct()

protocol.load(
    """
enum AttackType : uint8 {
    ATK_OPT_DPORT =  7,
    ATK_OPT_DOMAIN  =  8,
    ATK_OPT_NUM_SOCKETS =  24,
};

struct AttackTarget {
    DWORD   ipv4;
    BYTE    netmask;
};

struct AttackOption {
    AttackType   type;
    uint8   value_length;
    char    value[value_length];
};

struct MiraiAttack {
    uint16  total_length;
    uint32  duration;
    uint8   attack_id;
    uint8   target_count;
    AttackTarget targets[target_count];
    uint8   num_opts;
    AttackOption attack_options[num_opts];
};
"""
)

protocol.endian = ">"

if __name__ == "__main__":
    data = b"\x000\x00\x00\x00d\n\x01\x08\x08\x08\x08 \x03\x08\x16http://www.example.com\x07\x0280\x18\x045000"

    record = protocol.MiraiAttack(data)
    print(record)
    print("--")

    for t in record.targets:
        print(f"TARGET: {socket.inet_ntoa(struct.pack('!L', t.ipv4))}/{t.netmask}")

    for o in record.attack_options:
        print(f"OPTION: {o.type} - {o.value}")

    dumpstruct(protocol.MiraiAttack, data)



from io import BytesIO

from dissect import cstruct


def test_align_struct():
    d = """
    struct test {
        uint32  a;  // 0x00
        uint64  b;  // 0x08
        uint16  c;  // 0x10
        uint32  d;  // 0x14
        uint8   e;  // 0x18
        uint16  f;  // 0x1a
    };
    """
    c = cstruct.cstruct()
    c.load(d, align=True)

    fields = c.test.fields
    assert c.test.align
    assert c.test.alignment == 8
    assert c.test.size == 32
    assert fields[0].offset == 0x00
    assert fields[1].offset == 0x08
    assert fields[2].offset == 0x10
    assert fields[3].offset == 0x14
    assert fields[4].offset == 0x18
    assert fields[5].offset == 0x1A

    buf = """
        00 00 00 00 00 00 00 00  08 00 00 00 00 00 00 00
        10 00 00 00 14 00 00 00  18 00 1a 00 00 00 00 00
    """
    buf = bytes.fromhex(buf)
    fh = BytesIO(buf)

    obj = c.test(fh)
    assert fh.tell() == 32

    for name, value in obj._values.items():
        assert c.test.lookup[name].offset == value

    assert obj.dumps() == buf


def test_align_union():
    d = """
    union test {
        uint32  a;
        uint64  b;
    };
    """
    c = cstruct.cstruct()
    c.load(d, align=True)

    assert c.test.align
    assert c.test.alignment == 8
    assert c.test.size == 8

    buf = """
        00 00 00 01 00 00 00 02
    """
    buf = bytes.fromhex(buf)
    fh = BytesIO(buf)

    obj = c.test(fh)
    assert fh.tell() == 8
    assert obj.a == 0x01000000
    assert obj.b == 0x0200000001000000

    assert obj.dumps() == buf


def test_align_array():
    d = """
    struct test {
        uint32  a;      // 0x00
        uint64  b[4];   // 0x08
        uint16  c;      // 0x28
        uint32  d[2];   // 0x2c
        uint64  e;      // 0x38
    };
    """
    c = cstruct.cstruct()
    c.load(d, align=True)

    fields = c.test.fields
    assert c.test.align
    assert c.test.alignment == 8
    assert c.test.size == 64
    assert fields[0].offset == 0x00
    assert fields[1].offset == 0x08
    assert fields[2].offset == 0x28
    assert fields[3].offset == 0x2C
    assert fields[4].offset == 0x38

    buf = """
        00 00 00 00 00 00 00 00  08 00 00 00 00 00 00 00
        10 00 00 00 00 00 00 00  18 00 00 00 00 00 00 00
        20 00 00 00 00 00 00 00  28 00 00 00 2c 00 00 00
        30 00 00 00 00 00 00 00  38 00 00 00 00 00 00 00
    """
    buf = bytes.fromhex(buf)

    obj = c.test(buf)

    assert obj.a == 0x00
    assert obj.b == [0x08, 0x10, 0x18, 0x20]
    assert obj.c == 0x28
    assert obj.d == [0x2C, 0x30]
    assert obj.e == 0x38

    assert obj.dumps() == buf


def test_align_struct_array():
    d = """
    struct test {
        uint32  a;      // 0x00
        uint64  b;      // 0x08
    };

    struct array {
        test    a[4];
    };
    """
    c = cstruct.cstruct()
    c.load(d, align=True)

    fields = c.test.fields
    assert c.test.align
    assert c.test.alignment == 8
    assert c.test.size == 16
    assert fields[0].offset == 0x00
    assert fields[1].offset == 0x08

    buf = """
        00 00 00 00 00 00 00 00  08 00 00 00 00 00 00 00
        10 00 00 00 00 00 00 00  18 00 00 00 00 00 00 00
        20 00 00 00 00 00 00 00  28 00 00 00 00 00 00 00
        30 00 00 00 00 00 00 00  38 00 00 00 00 00 00 00
    """
    buf = bytes.fromhex(buf)

    obj = c.array(buf)

    assert obj.a[0].a == 0x00
    assert obj.a[0].b == 0x08
    assert obj.a[1].a == 0x10
    assert obj.a[1].b == 0x18
    assert obj.a[2].a == 0x20
    assert obj.a[2].b == 0x28
    assert obj.a[3].a == 0x30
    assert obj.a[3].b == 0x38

    assert obj.dumps() == buf


def test_align_dynamic():
    d = """
    struct test {
        uint8   a;      // 0x00 (value is 6 in test case)
        uint16  b[a];   // 0x02
        uint32  c;      // 0x?? (0x10 in test case)
        uint64  d;      // 0x?? (0x18 in test case)
        uint8   e;      // 0x?? (0x20, value is 2 in test case)
        uint32  f[e];   // 0x?? (0x24 in test case)
        uint64  g;      // 0x?? (0x30 in test case)
    };
    """
    c = cstruct.cstruct()
    c.load(d, align=True)

    fields = c.test.fields
    assert fields[0].offset == 0
    assert fields[1].offset == 2
    assert fields[2].offset is None
    assert fields[3].offset is None
    assert fields[4].offset is None
    assert fields[5].offset is None
    assert fields[6].offset is None

    buf = """
        06 00 02 00 04 00 06 00  08 00 0a 00 0c 00 00 00
        10 00 00 00 00 00 00 00  18 00 00 00 00 00 00 00
        02 00 00 00 24 00 00 00  28 00 00 00 00 00 00 00
        30 00 00 00 00 00 00 00
    """
    buf = bytes.fromhex(buf)
    obj = c.test(buf)

    assert obj.a == 0x06
    assert obj.b == [0x02, 0x04, 0x06, 0x08, 0x0A, 0x0C]
    assert obj.c == 0x10
    assert obj.d == 0x18
    assert obj.e == 0x02
    assert obj.f == [0x24, 0x28]
    assert obj.g == 0x30

    assert obj.dumps() == buf


def test_align_nested_struct():
    d = """
    struct test {
        uint32  a;      // 0x00
        struct {
            uint64  b;  // 0x08
            uint32  c;  // 0x10
        } nested;
        uint64  d;      // 0x18
    };
    """
    c = cstruct.cstruct()
    c.load(d, align=True)

    fields = c.test.fields
    assert fields[0].offset == 0x00
    assert fields[1].offset == 0x08
    assert fields[2].offset == 0x18

    buf = """
        00 00 00 00 00 00 00 00  08 00 00 00 00 00 00 00
        10 00 00 00 00 00 00 00  18 00 00 00 00 00 00 00
    """
    buf = bytes.fromhex(buf)
    obj = c.test(buf)

    assert obj.a == 0x00
    assert obj.nested.b == 0x08
    assert obj.nested.c == 0x10
    assert obj.d == 0x18

    assert obj.dumps() == buf


def test_align_bitfield():
    d = """
    struct test {
        uint16  a:4;    // 0x00
        uint16  b:4;
        uint64  c:4;    // 0x08
        uint64  d:4;
        uint16  e;      // 0x10
        uint32  f:4;    // 0x14
        uint64  g;      // 0x18
    };
    """
    c = cstruct.cstruct()
    c.load(d, align=True)

    fields = c.test.fields
    assert fields[0].offset == 0x00
    assert fields[1].offset is None
    assert fields[2].offset == 0x08
    assert fields[3].offset is None
    assert fields[4].offset == 0x10
    assert fields[5].offset == 0x14
    assert fields[6].offset == 0x18

    buf = """
        12 00 00 00 00 00 00 00  12 00 00 00 00 00 00 00
        10 00 00 00 02 00 00 00  18 00 00 00 00 00 00 00
    """
    buf = bytes.fromhex(buf)
    obj = c.test(buf)

    assert obj.a == 0b10
    assert obj.b == 0b01
    assert obj.c == 0b10
    assert obj.d == 0b01
    assert obj.e == 0x10
    assert obj.f == 0b10
    assert obj.g == 0x18

    assert obj.dumps() == buf


def test_align_pointer():
    d = """
    struct test {
        uint32  a;
        uint32  *b;
        uint16  c;
        uint16  d;
    };
    """
    c = cstruct.cstruct()
    c.load(d, align=True)

    fields = c.test.fields
    assert c.test.align
    assert c.test.alignment == 8
    assert c.test.size == 24
    assert fields[0].offset == 0x00
    assert fields[1].offset == 0x08
    assert fields[2].offset == 0x10
    assert fields[3].offset == 0x12

    buf = """
        00 00 00 00 00 00 00 00  18 00 00 00 00 00 00 00
        10 00 12 00 00 00 00 00  18 00 00 00
    """
    buf = bytes.fromhex(buf)
    obj = c.test(buf)

    assert obj.a == 0x00
    assert obj.b.dereference() == 0x18
    assert obj.c == 0x10
    assert obj.d == 0x12

    assert obj.dumps() == buf[:-4]  # Without pointer value



import pytest

from dissect import cstruct
from dissect.cstruct.utils import dumpstruct, hexdump

from .utils import verify_compiled


def test_hexdump(capsys):
    hexdump(b"\x00" * 16)
    captured = capsys.readouterr()
    assert captured.out == "00000000  00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00   ................\n"

    out = hexdump(b"\x00" * 16, output="string")
    assert out == "00000000  00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00   ................"

    out = hexdump(b"\x00" * 16, output="generator")
    assert next(out) == "00000000  00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00   ................"

    with pytest.raises(ValueError) as excinfo:
        hexdump("b\x00", output="str")
    assert str(excinfo.value) == "Invalid output argument: 'str' (should be 'print', 'generator' or 'string')."


def test_dumpstruct(capsys, compiled):
    cdef = """
    struct test {
        uint32 testval;
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    buf = b"\x39\x05\x00\x00"
    obj = cs.test(buf)

    dumpstruct(cs.test, buf)
    captured_1 = capsys.readouterr()

    dumpstruct(obj)
    captured_2 = capsys.readouterr()

    assert captured_1.out == captured_2.out

    out_1 = dumpstruct(cs.test, buf, output="string")
    out_2 = dumpstruct(obj, output="string")

    assert out_1 == out_2

    with pytest.raises(ValueError) as excinfo:
        dumpstruct(obj, output="generator")
    assert str(excinfo.value) == "Invalid output argument: 'generator' (should be 'print' or 'string')."



from dissect.cstruct.types import RawType


class VoidType(RawType):
    """Implements a void type."""

    def __init__(self):
        super().__init__(None, "void")

    def _read(self, stream) -> None:
        return None



import pytest

from dissect import cstruct
from dissect.cstruct.types import BytesInteger

from .utils import verify_compiled


def test_bytesinteger_unsigned():
    cs = cstruct.cstruct()

    assert cs.uint24(b"AAA") == 0x414141
    assert cs.uint24(b"\xff\xff\xff") == 0xFFFFFF
    assert cs.uint24[4](b"AAABBBCCCDDD") == [0x414141, 0x424242, 0x434343, 0x444444]
    assert cs.uint48(b"AAAAAA") == 0x414141414141
    assert cs.uint48(b"\xff\xff\xff\xff\xff\xff") == 0xFFFFFFFFFFFF
    assert cs.uint48[4](b"AAAAAABBBBBBCCCCCCDDDDDD") == [0x414141414141, 0x424242424242, 0x434343434343, 0x444444444444]

    uint40 = BytesInteger(cs, "uint40", 5, signed=False)
    assert uint40(b"AAAAA") == 0x4141414141
    assert uint40(b"\xff\xff\xff\xff\xff") == 0xFFFFFFFFFF
    assert uint40[2](b"AAAAABBBBB") == [0x4141414141, 0x4242424242]
    assert uint40[None](b"AAAAA\x00") == [0x4141414141]


def test_bytesinteger_signed():
    cs = cstruct.cstruct()

    assert cs.int24(b"\xff\x00\x00") == 255
    assert cs.int24(b"\xff\xff\xff") == -1
    assert cs.int24[4](b"\xff\xff\xff\xfe\xff\xff\xfd\xff\xff\xfc\xff\xff") == [-1, -2, -3, -4]

    int40 = BytesInteger(cs, "int40", 5, signed=True)
    assert int40(b"AAAAA") == 0x4141414141
    assert int40(b"\xff\xff\xff\xff\xff") == -1
    assert int40[2](b"\xff\xff\xff\xff\xff\xfe\xff\xff\xff\xff") == [-1, -2]


def test_bytesinteger_unsigned_be():
    cs = cstruct.cstruct()
    cs.endian = ">"

    assert cs.uint24(b"\x00\x00\xff") == 255
    assert cs.uint24(b"\xff\xff\xff") == 0xFFFFFF
    assert cs.uint24[3](b"\x00\x00\xff\x00\x00\xfe\x00\x00\xfd") == [255, 254, 253]

    uint40 = BytesInteger(cs, "uint40", 5, signed=False)
    assert uint40(b"\x00\x00\x00\x00\xff") == 255
    assert uint40(b"\xff\xff\xff\xff\xff") == 0xFFFFFFFFFF
    assert uint40[2](b"\x00\x00\x00\x00A\x00\x00\x00\x00B") == [0x41, 0x42]


def test_bytesinteger_signed_be():
    cs = cstruct.cstruct()
    cs.endian = ">"

    assert cs.int24(b"\x00\x00\xff") == 255
    assert cs.int24(b"\xff\xff\x01") == -255
    assert cs.int24[3](b"\xff\xff\x01\xff\xff\x02\xff\xff\x03") == [-255, -254, -253]

    int40 = BytesInteger(cs, "int40", 5, signed=True)
    assert int40(b"\x00\x00\x00\x00\xff") == 255
    assert int40(b"\xff\xff\xff\xff\xff") == -1
    assert int40(b"\xff\xff\xff\xff\x01") == -255
    assert int40[2](b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xfe") == [-1, -2]


def test_bytesinteger_struct_signed(compiled):
    cdef = """
    struct test {
        int24   a;
        int24   b[2];
        int24   len;
        int24   dync[len];
        int24   c;
        int24   d[3];
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    buf = b"AAABBBCCC\x02\x00\x00DDDEEE\xff\xff\xff\x01\xff\xff\x02\xff\xff\x03\xff\xff"
    obj = cs.test(buf)

    assert obj.a == 0x414141
    assert obj.b == [0x424242, 0x434343]
    assert obj.len == 0x02
    assert obj.dync == [0x444444, 0x454545]
    assert obj.c == -1
    assert obj.d == [-255, -254, -253]
    assert obj.dumps() == buf


def test_bytesinteger_struct_unsigned(compiled):
    cdef = """
    struct test {
        uint24  a;
        uint24  b[2];
        uint24  len;
        uint24  dync[len];
        uint24  c;
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    buf = b"AAABBBCCC\x02\x00\x00DDDEEE\xff\xff\xff"
    obj = cs.test(buf)

    assert obj.a == 0x414141
    assert obj.b == [0x424242, 0x434343]
    assert obj.len == 0x02
    assert obj.dync == [0x444444, 0x454545]
    assert obj.c == 0xFFFFFF
    assert obj.dumps() == buf


def test_bytesinteger_struct_signed_be(compiled):
    cdef = """
    struct test {
        int24   a;
        int24   b[2];
        int24   len;
        int24   dync[len];
        int24   c;
        int24   d[3];
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)
    cs.endian = ">"

    assert verify_compiled(cs.test, compiled)

    buf = b"AAABBBCCC\x00\x00\x02DDDEEE\xff\xff\xff\xff\xff\x01\xff\xff\x02\xff\xff\x03"
    obj = cs.test(buf)

    assert obj.a == 0x414141
    assert obj.b == [0x424242, 0x434343]
    assert obj.len == 0x02
    assert obj.dync == [0x444444, 0x454545]
    assert obj.c == -1
    assert obj.d == [-255, -254, -253]
    assert obj.dumps() == buf


def test_bytesinteger_struct_unsigned_be(compiled):
    cdef = """
    struct test {
        uint24  a;
        uint24  b[2];
        uint24  len;
        uint24  dync[len];
        uint24  c;
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)
    cs.endian = ">"

    assert verify_compiled(cs.test, compiled)

    buf = b"AAABBBCCC\x00\x00\x02DDDEEE\xff\xff\xff"
    obj = cs.test(buf)

    assert obj.a == 0x414141
    assert obj.b == [0x424242, 0x434343]
    assert obj.len == 0x02
    assert obj.dync == [0x444444, 0x454545]
    assert obj.c == 0xFFFFFF
    assert obj.dumps() == buf


def test_bytesinteger_range():
    cs = cstruct.cstruct()
    int8 = BytesInteger(cs, "int8", 1, signed=True)
    uint8 = BytesInteger(cs, "uint8", 1, signed=False)
    int16 = BytesInteger(cs, "int16", 2, signed=True)
    int24 = BytesInteger(cs, "int24", 3, signed=True)
    int8.dumps(127)
    int8.dumps(-128)
    uint8.dumps(255)
    uint8.dumps(0)
    int16.dumps(-32768)
    int16.dumps(32767)
    int24.dumps(-8388608)
    int24.dumps(8388607)
    with pytest.raises(OverflowError):
        int8.dumps(-129)
    with pytest.raises(OverflowError):
        int8.dumps(128)
    with pytest.raises(OverflowError):
        uint8.dumps(-1)
    with pytest.raises(OverflowError):
        uint8.dumps(256)
    with pytest.raises(OverflowError):
        int16.dumps(-32769)
    with pytest.raises(OverflowError):
        int16.dumps(32768)
    with pytest.raises(OverflowError):
        int24.dumps(-8388609)
    with pytest.raises(OverflowError):
        int24.dumps(8388608)



from __future__ import annotations

from io import BytesIO
from typing import Any, BinaryIO, List, TYPE_CHECKING

from dissect.cstruct.expression import Expression

if TYPE_CHECKING:
    from dissect.cstruct import cstruct


class BaseType:
    """Base class for cstruct type classes."""

    def __init__(self, cstruct: cstruct):
        self.cstruct = cstruct

    def __getitem__(self, count: int) -> Array:
        return Array(self.cstruct, self, count)

    def __call__(self, *args, **kwargs) -> Any:
        if len(args) > 0:
            return self.read(*args, **kwargs)

        result = self.default()
        if kwargs:
            for k, v in kwargs.items():
                setattr(result, k, v)

        return result

    def reads(self, data: bytes) -> Any:
        """Parse the given data according to the type that implements this class.

        Args:
            data: Byte string to parse.

        Returns:
            The parsed value of this type.
        """

        return self._read(BytesIO(data))

    def dumps(self, data: Any) -> bytes:
        """Dump the given data according to the type that implements this class.

        Args:
            data: Data to dump.

        Returns:
            The resulting bytes.
        """
        out = BytesIO()
        self._write(out, data)
        return out.getvalue()

    def read(self, obj: BinaryIO, *args, **kwargs) -> Any:
        """Parse the given data according to the type that implements this class.

        Args:
            obj: Data to parse. Can be a (byte) string or a file-like object.

        Returns:
            The parsed value of this type.
        """
        if isinstance(obj, (bytes, memoryview, bytearray)):
            return self.reads(obj)

        return self._read(obj)

    def write(self, stream: BinaryIO, data: Any) -> int:
        """Write the given data to a writable file-like object according to the
        type that implements this class.

        Args:
            stream: Writable file-like object to write to.
            data: Data to write.

        Returns:
            The amount of bytes written.
        """
        return self._write(stream, data)

    def _read(self, stream: BinaryIO) -> Any:
        raise NotImplementedError()

    def _read_array(self, stream: BinaryIO, count: int) -> List[Any]:
        return [self._read(stream) for _ in range(count)]

    def _read_0(self, stream: BinaryIO) -> List[Any]:
        raise NotImplementedError()

    def _write(self, stream: BinaryIO, data: Any) -> int:
        raise NotImplementedError()

    def _write_array(self, stream: BinaryIO, data: Any) -> int:
        num = 0
        for i in data:
            num += self._write(stream, i)

        return num

    def _write_0(self, stream: BinaryIO, data: Any) -> int:
        raise NotImplementedError()

    def default(self) -> Any:
        """Return a default value of this type."""
        raise NotImplementedError()

    def default_array(self, count: int) -> List[Any]:
        """Return a default array of this type."""
        return [self.default() for _ in range(count)]


class Array(BaseType):
    """Implements a fixed or dynamically sized array type.

    Example:
        When using the default C-style parser, the following syntax is supported:

            x[3] -> 3 -> static length.
            x[] -> None -> null-terminated.
            x[expr] -> expr -> dynamic length.
    """

    def __init__(self, cstruct: cstruct, type_: BaseType, count: int):
        self.type = type_
        self.count = count
        self.null_terminated = self.count is None
        self.dynamic = isinstance(self.count, Expression)
        self.alignment = type_.alignment
        super().__init__(cstruct)

    def __repr__(self) -> str:
        if self.null_terminated:
            return f"{self.type}[]"

        return f"{self.type}[{self.count}]"

    def __len__(self) -> int:
        if self.dynamic or self.null_terminated:
            raise TypeError("Dynamic size")

        return len(self.type) * self.count

    def _read(self, stream: BinaryIO, context: dict = None) -> List[Any]:
        if self.null_terminated:
            return self.type._read_0(stream)

        if self.dynamic:
            count = self.count.evaluate(context)
        else:
            count = self.count

        return self.type._read_array(stream, max(0, count))

    def _write(self, stream: BinaryIO, data: List[Any]) -> int:
        if self.null_terminated:
            return self.type._write_0(stream, data)

        return self.type._write_array(stream, data)

    def default(self) -> List[Any]:
        count = 0 if self.dynamic or self.null_terminated else self.count
        return self.type.default_array(count)


class RawType(BaseType):
    """Base class for raw types that have a name and size."""

    def __init__(self, cstruct: cstruct, name: str = None, size: int = 0, alignment: int = None):
        self.name = name
        self.size = size
        self.alignment = alignment or size
        super().__init__(cstruct)

    def __len__(self) -> int:
        return self.size

    def __repr__(self) -> str:
        if self.name:
            return self.name

        return BaseType.__repr__(self)

    def _read(self, stream: BinaryIO) -> Any:
        raise NotImplementedError()

    def _read_0(self, stream: BinaryIO) -> List[Any]:
        raise NotImplementedError()

    def _write(self, stream: BinaryIO, data: Any) -> int:
        raise NotImplementedError()

    def _write_0(self, stream: BinaryIO, data: List[Any]) -> int:
        raise NotImplementedError()

    def default(self) -> Any:
        raise NotImplementedError()



import pytest

from dissect import cstruct

from .utils import verify_compiled


def test_enum(compiled):
    cdef = """
    enum Test16 : uint16 {
        A = 0x1,
        B = 0x2
    };

    enum Test24 : uint24 {
        A = 0x1,    // comment, best one
        B = 0x2
    };

    enum Test32 : uint32 {
        A = 0x1,
        B = 0x2     // comment
    };

    struct test {
        Test16  a16;
        Test16  b16;
        Test24  a24;
        Test24  b24;
        Test32  a32;
        Test32  b32;        // this is a comment, awesome
        Test16  l[2];
    };

    struct test_term {
        Test16  null[];
    };

    struct test_expr {
        uint16  size;
        Test16  expr[size * 2];
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)
    assert verify_compiled(cs.test_term, compiled)
    assert verify_compiled(cs.test_expr, compiled)

    buf = b"\x01\x00\x02\x00\x01\x00\x00\x02\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x01\x00\x02\x00"
    obj = cs.test(buf)

    assert obj.a16.enum == cs.Test16 and obj.a16 == cs.Test16.A
    assert obj.b16.enum == cs.Test16 and obj.b16 == cs.Test16.B
    assert obj.a24.enum == cs.Test24 and obj.a24 == cs.Test24.A
    assert obj.b24.enum == cs.Test24 and obj.b24 == cs.Test24.B
    assert obj.a32.enum == cs.Test32 and obj.a32 == cs.Test32.A
    assert obj.b32.enum == cs.Test32 and obj.b32 == cs.Test32.B

    assert len(obj.l) == 2
    assert obj.l[0].enum == cs.Test16 and obj.l[0] == cs.Test16.A
    assert obj.l[1].enum == cs.Test16 and obj.l[1] == cs.Test16.B

    assert "A" in cs.Test16
    assert "Foo" not in cs.Test16
    assert cs.Test16(1) == cs.Test16["A"]
    assert cs.Test24(2) == cs.Test24.B
    assert cs.Test16.A != cs.Test24.A

    with pytest.raises(KeyError):
        cs.Test16["C"]

    with pytest.raises(AttributeError):
        cs.Test16.C

    assert obj.dumps() == buf

    buf = b"\x01\x00\x02\x00\x00\x00"
    assert cs.test_term(buf).null == [cs.Test16.A, cs.Test16.B]
    assert cs.test_term(null=[cs.Test16.A, cs.Test16.B]).dumps() == buf

    buf = b"\x01\x00\x01\x00\x02\x00"
    assert cs.test_expr(buf).expr == [cs.Test16.A, cs.Test16.B]
    assert cs.test_expr(size=1, expr=[cs.Test16.A, cs.Test16.B]).dumps() == buf

    obj = {
        cs.Test16.A: "Test16.A",
        cs.Test16.B: "Test16.B",
        cs.Test24.A: "Test24.A",
        cs.Test24.B: "Test24.B",
    }

    assert obj[cs.Test16.A] == "Test16.A"
    assert obj[cs.Test16(2)] == "Test16.B"
    assert obj[cs.Test24(1)] == "Test24.A"
    assert obj[cs.Test24.B] == "Test24.B"

    with pytest.raises(KeyError):
        obj[cs.Test32.A]


def test_enum_comments():
    cdef = """
    enum Inline { hello=7, world, foo, bar }; // inline enum

    enum Test {
        a = 2,  // comment, 2
        b,      // comment, 3
        c       // comment, 4
    };

    enum Odd {
        a = 0,          // hello, world
        b,              // next
        c,              // next
        d = 5, e, f     // inline, from 5
        g               // next
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef)

    assert cs.Inline.hello == 7
    assert cs.Inline.world == 8
    assert cs.Inline.foo == 9
    assert cs.Inline.bar == 10

    assert cs.Test.a == 2
    assert cs.Test.b == 3
    assert cs.Test.c == 4

    assert cs.Odd.a == 0
    assert cs.Odd.b == 1
    assert cs.Odd.c == 2

    assert cs.Odd.d == 5
    assert cs.Odd.e == 6
    assert cs.Odd.f == 7
    assert cs.Odd.g == 8

    assert cs.Test.a == cs.Test.a
    assert cs.Test.a != cs.Test.b


def test_enum_name(compiled):
    cdef = """
    enum Color: uint16 {
          RED = 1,
          GREEN = 2,
          BLUE = 3,
    };

    struct Pixel {
        uint8 x;
        uint8 y;
        Color color;
        uint32 hue;
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.Pixel, compiled)

    Color = cs.Color
    Pixel = cs.Pixel

    pixel = Pixel(b"\xFF\x0A\x01\x00\xAA\xBB\xCC\xDD")
    assert pixel.x == 255
    assert pixel.y == 10
    assert pixel.color.name == "RED"
    assert pixel.color.value == Color.RED
    assert pixel.color.value == 1
    assert pixel.hue == 0xDDCCBBAA

    # unknown enum values default to <enum name>_<value>
    pixel = Pixel(b"\x00\x00\xFF\x00\xAA\xBB\xCC\xDD")
    assert pixel.color.name == "Color_255"
    assert pixel.color.value == 0xFF


def test_enum_write(compiled):
    cdef = """
    enum Test16 : uint16 {
        A = 0x1,
        B = 0x2
    };

    enum Test24 : uint24 {
        A = 0x1,
        B = 0x2
    };

    enum Test32 : uint32 {
        A = 0x1,
        B = 0x2
    };

    struct test {
        Test16  a16;
        Test16  b16;
        Test24  a24;
        Test24  b24;
        Test32  a32;
        Test32  b32;
        Test16  list[2];
    };
    """
    cs = cstruct.cstruct()
    cs.load(cdef, compiled=compiled)

    assert verify_compiled(cs.test, compiled)

    obj = cs.test()
    obj.a16 = cs.Test16.A
    obj.b16 = cs.Test16.B
    obj.a24 = cs.Test24.A
    obj.b24 = cs.Test24.B
    obj.a32 = cs.Test32.A
    obj.b32 = cs.Test32.B
    obj.list = [cs.Test16.A, cs.Test16.B]

    assert obj.dumps() == b"\x01\x00\x02\x00\x01\x00\x00\x02\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x01\x00\x02\x00"



from __future__ import print_function

import ctypes as _ctypes
import sys
from typing import Any, BinaryIO

from dissect.cstruct.exceptions import ResolveError
from dissect.cstruct.parser import CStyleParser, TokenParser
from dissect.cstruct.types import (
    Array,
    BaseType,
    BytesInteger,
    CharType,
    Structure,
    PackedType,
    Pointer,
    VoidType,
    WcharType,
)


class cstruct:
    """Main class of cstruct. All types are registered in here.

    Args:
        endian: The endianness to use when parsing.
        pointer: The pointer type to use for Pointers.
    """

    DEF_CSTYLE = 1
    DEF_LEGACY = 2

    def __init__(self, endian: str = "<", pointer: str = None):
        self.endian = endian

        self.consts = {}
        self.lookups = {}
        # fmt: off
        self.typedefs = {
            # Internal types
            "int8": PackedType(self, "int8", 1, "b"),
            "uint8": PackedType(self, "uint8", 1, "B"),
            "int16": PackedType(self, "int16", 2, "h"),
            "uint16": PackedType(self, "uint16", 2, "H"),
            "int32": PackedType(self, "int32", 4, "i"),
            "uint32": PackedType(self, "uint32", 4, "I"),
            "int64": PackedType(self, "int64", 8, "q"),
            "uint64": PackedType(self, "uint64", 8, "Q"),
            "float": PackedType(self, "float", 4, "f"),
            "double": PackedType(self, "double", 8, "d"),
            "char": CharType(self),
            "wchar": WcharType(self),

            "int24": BytesInteger(self, "int24", 3, True, alignment=4),
            "uint24": BytesInteger(self, "uint24", 3, False, alignment=4),
            "int48": BytesInteger(self, "int48", 6, True, alignment=8),
            "uint48": BytesInteger(self, "uint48", 6, False, alignment=8),

            "void": VoidType(),

            # Common C types not covered by internal types
            "signed char": "int8",
            "unsigned char": "char",
            "short": "int16",
            "signed short": "short",
            "unsigned short": "uint16",
            "int": "int32",
            "signed int": "int",
            "unsigned int": "uint32",
            "long": "int32",
            "signed long": "long",
            "unsigned long": "uint32",
            "long long": "int64",
            "signed long long": "long long",
            "unsigned long long": "uint64",

            # Windows types
            "BYTE": "uint8",
            "CHAR": "char",
            "SHORT": "int16",
            "WORD": "uint16",
            "DWORD": "uint32",
            "LONG": "int32",
            "LONG32": "int32",
            "LONG64": "int64",
            "LONGLONG": "int64",
            "QWORD": "uint64",
            "WCHAR": "wchar",

            "UCHAR": "uint8",
            "USHORT": "uint16",
            "ULONG": "uint32",
            "ULONG64": "uint64",
            "ULONGLONG": "uint64",

            "INT": "int",
            "INT8": "int8",
            "INT16": "int16",
            "INT32": "int32",
            "INT64": "int64",

            "UINT": "uint",
            "UINT8": "uint8",
            "UINT16": "uint16",
            "UINT32": "uint32",
            "UINT64": "uint64",

            "__int8": "int8",
            "__int16": "int16",
            "__int32": "int32",
            "__int64": "int64",

            "wchar_t": "wchar",

            # GNU C types
            "int8_t": "int8",
            "int16_t": "int16",
            "int32_t": "int32",
            "int64_t": "int64",

            "uint8_t": "uint8",
            "uint16_t": "uint16",
            "uint32_t": "uint32",
            "uint64_t": "uint64",

            # Other convenience types
            "u1": "uint8",
            "u2": "uint16",
            "u4": "uint32",
            "u8": "uint64",
            "uchar": "uint8",
            "ushort": "unsigned short",
            "uint": "unsigned int",
            "ulong": "unsigned long",
        }
        # fmt: on

        pointer = pointer or "uint64" if sys.maxsize > 2**32 else "uint32"
        self.pointer = self.resolve(pointer)
        self._anonymous_count = 0

    def __getattr__(self, attr: str) -> Any:
        try:
            return self.resolve(self.typedefs[attr])
        except KeyError:
            pass

        try:
            return self.consts[attr]
        except KeyError:
            pass

        raise AttributeError(f"Invalid attribute: {attr}")

    def _next_anonymous(self) -> str:
        name = f"anonymous_{self._anonymous_count}"
        self._anonymous_count += 1
        return name

    def addtype(self, name: str, type_: BaseType, replace: bool = False) -> None:
        """Add a type or type reference.

        Args:
            name: Name of the type to be added.
            type_: The type to be added. Can be a str reference to another type
                or a compatible type class.

        Raises:
            ValueError: If the type already exists.
        """
        if not replace and (name in self.typedefs and self.resolve(self.typedefs[name]) != self.resolve(type_)):
            raise ValueError(f"Duplicate type: {name}")

        self.typedefs[name] = type_

    def load(self, definition: str, deftype: int = None, **kwargs) -> None:
        """Parse structures from the given definitions using the given definition type.

        Definitions can be parsed using different parsers. Currently, there's
        only one supported parser - DEF_CSTYLE. Parsers can add types and
        modify this cstruct instance. Arguments can be passed to parsers
        using kwargs.

        The CSTYLE parser was recently replaced with token based parser,
        instead of a strictly regex based one. The old parser is still available
        by using DEF_LEGACY.

        Args:
            definition: The definition to parse.
            deftype: The definition type to parse the definitions with.
            **kwargs: Keyword arguments for parsers.
        """
        deftype = deftype or cstruct.DEF_CSTYLE

        if deftype == cstruct.DEF_CSTYLE:
            TokenParser(self, **kwargs).parse(definition)
        elif deftype == cstruct.DEF_LEGACY:
            CStyleParser(self, **kwargs).parse(definition)

    def loadfile(self, path: str, deftype: int = None, **kwargs) -> None:
        """Load structure definitions from a file.

        The given path will be read and parsed using the .load() function.

        Args:
            path: The path to load definitions from.
            deftype: The definition type to parse the definitions with.
            **kwargs: Keyword arguments for parsers.
        """
        with open(path) as fh:
            self.load(fh.read(), deftype, **kwargs)

    def read(self, name: str, stream: BinaryIO) -> Any:
        """Parse data using a given type.

        Args:
            name: Type name to read.
            stream: File-like object or byte string to parse.

        Returns:
            The parsed data.
        """
        return self.resolve(name).read(stream)

    def resolve(self, name: str) -> BaseType:
        """Resolve a type name to get the actual type object.

        Types can be referenced using different names. When we want
        the actual type object, we need to resolve these references.

        Args:
            name: Type name to resolve.

        Returns:
            The resolved type object.

        Raises:
            ResolveError: If the type can't be resolved.
        """
        type_name = name
        if not isinstance(type_name, str):
            return type_name

        for _ in range(10):
            if type_name not in self.typedefs:
                raise ResolveError(f"Unknown type {name}")

            type_name = self.typedefs[type_name]

            if not isinstance(type_name, str):
                return type_name

        raise ResolveError(f"Recursion limit exceeded while resolving type {name}")


def ctypes(structure: Structure) -> _ctypes.Structure:
    """Create ctypes structures from cstruct structures."""
    fields = []
    for field in structure.fields:
        t = ctypes_type(field.type)
        fields.append((field.name, t))

    tt = type(structure.name, (_ctypes.Structure,), {"_fields_": fields})
    return tt


def ctypes_type(type_: BaseType) -> Any:
    mapping = {
        "I": _ctypes.c_ulong,
        "i": _ctypes.c_long,
        "b": _ctypes.c_int8,
    }

    if isinstance(type_, PackedType):
        return mapping[type_.packchar]

    if isinstance(type_, CharType):
        return _ctypes.c_char

    if isinstance(type_, Array):
        subtype = ctypes_type(type_.type)
        return subtype * type_.count

    if isinstance(type_, Pointer):
        subtype = ctypes_type(type_.type)
        return _ctypes.POINTER(subtype)

    raise NotImplementedError(f"Type not implemented: {type_.__class__.__name__}")



from __future__ import annotations

from typing import BinaryIO, TYPE_CHECKING

from dissect.cstruct.types import RawType

if TYPE_CHECKING:
    from dissect.cstruct import cstruct


class CharType(RawType):
    """Implements a character type that can properly handle strings."""

    def __init__(self, cstruct: cstruct):
        super().__init__(cstruct, "char", size=1, alignment=1)

    def _read(self, stream: BinaryIO) -> bytes:
        return stream.read(1)

    def _read_array(self, stream: BinaryIO, count: int) -> bytes:
        if count == 0:
            return b""

        return stream.read(count)

    def _read_0(self, stream: BinaryIO) -> bytes:
        byte_array = []
        while True:
            bytes_stream = stream.read(1)
            if bytes_stream == b"":
                raise EOFError()

            if bytes_stream == b"\x00":
                break

            byte_array.append(bytes_stream)

        return b"".join(byte_array)

    def _write(self, stream: BinaryIO, data: bytes) -> int:
        if isinstance(data, int):
            data = chr(data)

        if isinstance(data, str):
            data = data.encode("latin-1")

        return stream.write(data)

    def _write_array(self, stream: BinaryIO, data: bytes) -> int:
        return self._write(stream, data)

    def _write_0(self, stream: BinaryIO, data: bytes) -> int:
        return self._write(stream, data + b"\x00")

    def default(self) -> bytes:
        return b"\x00"

    def default_array(self, count: int) -> bytes:
        return b"\x00" * count
